# cython: language_level=3str
cimport cython

from ._event cimport Event


@cython.final
cdef class Query:
    cdef:
        readonly tuple query_contains
        readonly tuple query_endswith
        readonly tuple query_eq
        readonly tuple query_gt
        readonly tuple query_gte
        readonly tuple query_in
        readonly tuple query_lt
        readonly tuple query_lte
        readonly tuple query_regex
        readonly tuple query_startswith


@cython.final
cdef class And:
    cdef:
        readonly tuple predicates


@cython.final
cdef class Or:
    cdef:
        readonly tuple predicates


@cython.final
cdef class Not:
    cdef:
        readonly object predicate


@cython.final
cdef class When:
    cdef:
        readonly object condition
        readonly tuple actions


@cython.final
cdef class From:
    cdef:
        readonly object condition
        readonly object predicate
        readonly int watermark
        readonly int origin_depth
        readonly int origin_calls


@cython.final
cdef class Backlog:
    cdef:
        readonly object condition
        readonly int size
        readonly int stack
        readonly bint vars
        readonly bint strip
        readonly object action
        readonly object _try_repr
        readonly object _filter
        readonly object queue

cdef fast_And_call(And self, Event event)
cdef fast_From_call(From self, Event event)
cdef fast_Not_call(Not self, Event event)
cdef fast_Or_call(Or self, Event event)
cdef fast_Query_call(Query self, Event event)
cdef fast_When_call(When self, Event event)
cdef fast_call(object callable, Event event)



import atexit
import functools
import inspect
import os
import sys
import warnings
import weakref

from .actions import Action
from .actions import CallPrinter
from .actions import CodePrinter
from .actions import Debugger
from .actions import ErrorSnooper
from .actions import Manhole
from .actions import StackPrinter
from .actions import VarsPrinter
from .actions import VarsSnooper

try:
    if os.environ.get('PUREPYTHONHUNTER'):
        raise ImportError('Cython speedups are disabled.')

    from ._event import Event
    from ._predicates import And as _And
    from ._predicates import Backlog as _Backlog
    from ._predicates import From as _From
    from ._predicates import Not as _Not
    from ._predicates import Or as _Or
    from ._predicates import Query
    from ._predicates import When
    from ._tracer import Tracer
except ImportError:
    from .event import Event  # noqa
    from .predicates import And as _And
    from .predicates import Backlog as _Backlog
    from .predicates import From as _From
    from .predicates import Not as _Not
    from .predicates import Or as _Or
    from .predicates import Query
    from .predicates import When
    from .tracer import Tracer

try:
    from ._version import version as __version__
except ImportError:
    __version__ = '3.8.0'

__all__ = (
    'And',
    'Backlog',
    'CallPrinter',
    'CodePrinter',
    'Debugger',
    'ErrorSnooper',
    'From',
    'Manhole',
    'Not',
    'Or',
    'Q',
    'Query',
    'StackPrinter',
    'VarsPrinter',
    'VarsSnooper',
    'When',
    'stop',
    'trace',
)
THREADING_SUPPORT_ALIASES = (
    'threading_support',
    'threads_support',
    'thread_support',
    'threadingsupport',
    'threadssupport',
    'threadsupport',
    'threading',
    'threads',
    'thread',
)
TRACER_OPTION_NAMES = (*THREADING_SUPPORT_ALIASES, 'clear_env_var', 'profile')
_last_tracer = None
_default_trace_args = None
_default_config = {}
_default_stream = sys.stderr


def _embed_via_environment():
    if 'PYTHONHUNTER' in os.environ:
        try:
            eval(
                f'trace({os.environ["PYTHONHUNTER"]})',
                {
                    'And': And,
                    'Backlog': Backlog,
                    'CallPrinter': CallPrinter,
                    'CodePrinter': CodePrinter,
                    'Debugger': Debugger,
                    'ErrorSnooper': ErrorSnooper,
                    'From': From,
                    'Manhole': Manhole,
                    'Not': Not,
                    'Or': Or,
                    'Q': Q,
                    'Query': Query,
                    'StackPrinter': StackPrinter,
                    'VarsPrinter': VarsPrinter,
                    'VarsSnooper': VarsSnooper,
                    'When': When,
                    'trace': trace,
                },
            )
        except Exception as exc:
            sys.stderr.write(f'Failed to load hunter from PYTHONHUNTER environ variable {os.environ["PYTHONHUNTER"]!r}: {exc!r}\n')


def _prepare_config(*args, **kwargs):
    _default_config.clear()
    _default_config.update((key.lower(), val) for key, val in kwargs.items())
    options = {}
    predicates = []

    for key, value in list(_default_config.items()):
        if key in TRACER_OPTION_NAMES:
            options[key] = _default_config.pop(key)
            continue
        elif key in (
            # builtin actions config
            'klass',
            'stream',
            'force_colors',
            'force_pid',
            'filename_alignment',
            'thread_alignment',
            'pid_alignment',
            'repr_limit',
            'repr_func',
        ):
            continue

        try:
            Q(**{key: value})
        except TypeError:
            pass
        else:
            options[key] = _default_config.pop(key)
            continue

        _default_config.pop(key)
        sys.stderr.write(f'Discarded config from PYTHONHUNTERCONFIG {key}={value!r}: Unknown option\n')
    for position, predicate in enumerate(args):
        if callable(predicate):
            predicates.append(predicate)
        else:
            sys.stderr.write(f'Discarded config from PYTHONHUNTERCONFIG {predicate} (position {position}): Not a callable\n')

    return predicates, options


def Q(*predicates, **query):
    """
    Helper that handles situations where :class:`hunter.predicates.Query` objects (or other callables)
    are passed in as positional arguments - it conveniently converts those to a
    :class:`hunter.predicates.And` predicate.

    See Also:
         :class:`hunter.predicates.Query`
    """
    optional_actions = query.pop('actions', [])
    if 'action' in query:
        optional_actions.append(query.pop('action'))

    for p in predicates:
        if not callable(p):
            raise TypeError(f'Predicate {p!r} is not callable.')

    for a in optional_actions:
        if not callable(a):
            raise TypeError(f'Action {a!r} is not callable.')

    if predicates:
        predicates = tuple(p() if inspect.isclass(p) and issubclass(p, Action) else p for p in predicates)
        if any(isinstance(p, (CallPrinter, CodePrinter)) for p in predicates):
            # the user provided an action as a filter, remove the action then to prevent double output
            for action in optional_actions:
                if action in (CallPrinter, CodePrinter) or isinstance(action, (CallPrinter, CodePrinter)):
                    optional_actions.remove(action)
        if query:
            predicates += (Query(**query),)

        result = And(*predicates)
    else:
        result = Query(**query)

    if optional_actions:
        result = When(result, *optional_actions)

    return result


def _merge(*predicates, **query):
    if predicates:
        if query:
            predicates += (Query(**query),)
        return And(*predicates)
    else:
        return Query(**query)


def _flatten(cls, predicate, *predicates):
    if not predicates:
        return predicate
    else:
        all_predicates = []
        if isinstance(predicate, cls):
            all_predicates.extend(predicate.predicates)
        else:
            all_predicates.append(predicate)

        for p in predicates:
            if isinstance(p, cls):
                all_predicates.extend(p.predicates)
            else:
                all_predicates.append(p)
        return cls(*all_predicates)


def And(*predicates, **kwargs):
    """
    Helper that flattens out ``predicates`` in a single :class:`hunter.predicates.And` object if possible.
    As a convenience it converts ``kwargs`` to a single :class:`hunter.predicates.Query` instance.

    Args:
        *predicates (callables): Callables that returns True/False or :class:`hunter.predicates.Query` objects.
        **kwargs: Arguments that may be passed to :class:`hunter.predicates.Query`.

    Returns: A :class:`hunter.predicates.And` instance.

    See Also:
         :class:`hunter.predicates.And`
    """
    if kwargs:
        predicates += (Query(**kwargs),)
    return _flatten(_And, *predicates)


def Or(*predicates, **kwargs):
    """
    Helper that flattens out ``predicates`` in a single :class:`hunter.predicates.Or` object if possible.
    As a convenience it converts ``kwargs`` to multiple :class:`hunter.predicates.Query` instances.

    Args:
        *predicates (callables): Callables that returns True/False or :class:`hunter.predicates.Query` objects.
        **kwargs: Arguments that may be passed to :class:`hunter.predicates.Query`.

    Returns: A :class:`hunter.predicates.Or` instance.

    See Also:
         :class:`hunter.predicates.Or`
    """
    if kwargs:
        predicates += tuple(Query(**{key: value}) for key, value in kwargs.items())
    return _flatten(_Or, *predicates)


def Not(*predicates, **kwargs):
    """
    Helper that flattens out ``predicates`` in a single :class:`hunter.predicates.And` object if possible.
    As a convenience it converts ``kwargs`` to multiple :class:`hunter.predicates.Query` instances.

    Args:
        *predicates (callables): Callables that returns True/False or :class:`hunter.predicates.Query` objects.
        **kwargs: Arguments that may be passed to :class:`hunter.predicates.Query`.

    Returns: A :class:`hunter.predicates.Not` instance (possibly containing a :class:`hunter.predicates.And` instance).

    See Also:
         :class:`hunter.predicates.Not`
    """
    if kwargs:
        predicates += (Query(**kwargs),)
    if len(predicates) > 1:
        return _Not(_flatten(_And, *predicates))
    else:
        return _Not(*predicates)


def From(condition=None, predicate=None, watermark=0, **kwargs):
    """
    Helper that converts keyword arguments to ``From(condition=Q(**normal_kwargs), predicate=Q(**rel_kwargs)``
    where ``rel_kwargs`` are all the kwargs that start with "depth" or "calls".

    Args:
        condition (callable): A callable that returns True/False or a :class:`hunter.predicates.Query` object.
        predicate (callable): Optional callable that returns True/False or a :class:`hunter.predicates.Query` object to
            run after ``condition`` first returns ``True``.
        **kwargs: Arguments that are passed to :func:`hunter.Q`. Any kwarg that starts with "depth" or "calls" will be included `predicate`.

    Examples:
        ``From(function='foobar', depth_lt=5)`` coverts to ``From(Q(function='foobar'), Q(depth_lt=5))``.
        The depth filter is moved in the predicate because it would not have any effect as a condition - it stop being called after it
        returns True, thus it doesn't have the intended effect (a limit to how deep to trace from ``foobar``).

    See Also:
         :class:`hunter.predicates.From`
    """
    if predicate is None and condition is None:
        condition_kwargs = {key: value for key, value in kwargs.items() if not key.startswith('depth') and not key.startswith('calls')}
        predicate_kwargs = {key: value for key, value in kwargs.items() if key.startswith(('depth', 'calls'))}
        return _From(Q(**condition_kwargs), Q(**predicate_kwargs), watermark)
    else:
        if kwargs:
            raise TypeError(f"Unexpected arguments {kwargs.keys()}. Don't combine positional with keyword arguments.")
        return _From(condition, predicate, watermark)


def Backlog(*conditions, **kwargs):
    """
    Helper that merges kwargs and conditions prior to creating the :class:`~hunter.predicates.Backlog`.

    Args:
        *conditions (callable): Optional :class:`~hunter.predicates.Query` object or a callable that returns True/False.
        size (int): Number of events that the backlog stores. Effectively this is the ``maxlen`` for the internal deque.
        stack (int): Stack size to fill. Setting this to ``0`` disables creating fake call events.
        vars (bool): Makes global/local variables available in the stored events.
            This is an expensive option - it will use ``action.try_repr`` on all the variables.
        strip (bool): If this option is set then the backlog will be cleared every time an event matching the ``condition`` is found.
            Disabling this may show more context every time an event matching the ``condition`` is found but said context may also be
            duplicated across multiple matches.
        action (ColorStreamAction): A ColorStreamAction to display the stored events when an event matching the ``condition`` is found.
        filter (callable): Optional :class:`~hunter.predicates.Query` object or a callable that returns True/False to filter the stored
            events with.
        **kwargs: Arguments that are passed to :func:`hunter.Q`. Any kwarg that starts with "depth" or "calls" will be included `predicate`.

    See Also:
         :class:`hunter.predicates.Backlog`
    """
    action = kwargs.pop('action', CallPrinter)
    filter = kwargs.pop('filter', None)
    size = kwargs.pop('size', 100)
    stack = kwargs.pop('stack', 10)
    strip = kwargs.pop('strip', True)
    vars = kwargs.pop('vars', False)
    if not conditions and not kwargs:
        raise TypeError(
            "Backlog needs at least 1 condition (it doesn't have any effect without one besides making everything incredibly slow)."
        )
    return _Backlog(
        _merge(*conditions, **kwargs),
        size=size,
        stack=stack,
        vars=vars,
        strip=strip,
        action=action,
        filter=filter,
    )


def stop():
    """
    Stop tracing. Restores previous tracer (if there was any).
    """
    global _last_tracer

    if _last_tracer is None:
        warnings.warn('There is no tracer to stop.', stacklevel=2)
    else:
        _last_tracer.stop()
        _last_tracer = None


class Stop(Action):
    def __call__(self, event):
        stop()


def _prepare_predicate(*predicates, **options):
    if 'action' not in options and 'actions' not in options:
        options['action'] = CallPrinter

    return Q(*predicates, **options)


def _apply_config(predicates, options):
    if _default_trace_args is None:
        return predicates, options
    else:
        config_predicates, config_options = _default_trace_args
        return predicates + tuple(config_predicates), dict(config_options, **options)


def trace(*predicates, **options):
    """
    Starts tracing. Can be used as a context manager (with slightly incorrect semantics - it starts tracing
    before ``__enter__`` is called).

    Parameters:
        *predicates (callables): Runs actions if **all** of the given predicates match.
    Keyword Args:
        clear_env_var: Disables tracing in subprocess. Default: ``False``.
        threading_support: Enable tracing *new* threads. Default: ``None``.

            Modes:

            - ``None`` - automatic (enabled but actions only prefix with thread name if more than 1 thread)
            - ``False`` - completely disabled
            - ``True`` - enabled (actions always prefix with thread name)

            You can also use:
            ``threads_support``, ``thread_support``, ``threadingsupport``, ``threadssupport``, ``threadsupport``,
            ``threading``, ``threads`` or ``thread``.
        action: Action to run if all the predicates return ``True``. Default: ``CodePrinter``.
        actions: Actions to run (in case you want more than 1).
        **kwargs: for convenience you can also pass anything that you'd pass to :obj:`hunter.Q`

    See Also:
         :class:`hunter.tracer.Tracer` or :class:`hunter.event.Event`
    """
    global _last_tracer

    predicates, options = _apply_config(predicates, options)

    clear_env_var = options.pop('clear_env_var', False)
    profiling_mode = options.pop('profile', False)
    threading_support = None
    for alias in THREADING_SUPPORT_ALIASES:
        if alias in options:
            threading_support = options.pop(alias)
    predicate = _prepare_predicate(*predicates, **options)

    if clear_env_var:
        os.environ.pop('PYTHONHUNTER', None)

    _last_tracer = Tracer(threading_support, profiling_mode)

    @atexit.register
    def atexit_cleanup(ref=weakref.ref(_last_tracer)):  # noqa: B008
        maybe_tracer = ref()
        if maybe_tracer is not None:
            maybe_tracer.stop()

    return _last_tracer.trace(predicate)


def wrap(function_to_trace=None, **trace_options):
    """
    Functions decorated with this will be traced.

    Use ``local=True`` to only trace local code, eg::

        @hunter.wrap(local=True)
        def my_function():
            ...

    Keyword arguments are allowed, eg::

        @hunter.wrap(action=hunter.CallPrinter)
        def my_function():
            ...

    Or, filters::

        @hunter.wrap(module='foobar')
        def my_function():
            ...
    """

    def tracing_decorator(func):
        predicates = []
        local = trace_options.pop('local', False)
        if local:
            predicates.append(Query(depth_lt=2))
        predicates.append(
            From(
                Query(kind='call', function=func.__name__),
                Not(
                    When(
                        Query(calls_gt=0, depth=0) & Not(Query(kind='return')),
                        Stop,
                    )
                ),
                watermark=-1,
            )
        )

        if inspect.isasyncgenfunction(func):

            @functools.wraps(func)
            async def tracing_wrapper(*args, **kwargs):
                local_tracer = trace(*predicates, **trace_options)
                try:
                    async for item in func(*args, **kwargs):
                        yield item
                finally:
                    local_tracer.stop()

        elif inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def tracing_wrapper(*args, **kwargs):
                local_tracer = trace(*predicates, **trace_options)
                try:
                    return await func(*args, **kwargs)
                finally:
                    local_tracer.stop()

        elif inspect.isgeneratorfunction(func):

            @functools.wraps(func)
            def tracing_wrapper(*args, **kwargs):
                local_tracer = trace(*predicates, **trace_options)
                try:
                    yield from func(*args, **kwargs)
                finally:
                    local_tracer.stop()

        else:

            @functools.wraps(func)
            def tracing_wrapper(*args, **kwargs):
                local_tracer = trace(*predicates, **trace_options)
                try:
                    return func(*args, **kwargs)
                finally:
                    local_tracer.stop()

        return tracing_wrapper

    if function_to_trace is None:
        return tracing_decorator
    else:
        return tracing_decorator(function_to_trace)


def load_config(*args, **kwargs):
    global _default_trace_args
    if args or kwargs:
        _default_trace_args = _prepare_config(*args, **kwargs)
    else:
        config = os.environ.get('PYTHONHUNTERCONFIG', '')
        try:
            _default_trace_args = eval(f'_prepare_config({config})')
        except Exception as exc:
            sys.stderr.write(f'Failed to load hunter config from PYTHONHUNTERCONFIG {config!r}: {exc!r}\n')
            _default_trace_args = (), ()


load_config()



# ruff: noqa: S103 S108 S603
import argparse
import errno
import json
import os
import platform
import signal
import socket
import sys
import time
from contextlib import closing
from contextlib import contextmanager
from subprocess import check_call

import hunter

if platform.system() == 'Windows':
    print('ERROR: This tool does not work on Windows.', file=sys.stderr)
    sys.exit(1)
else:
    import manhole
    from manhole import get_peercred
    from manhole.cli import parse_signal

__all__ = ('install',)


def install(**kwargs):
    kwargs.setdefault('oneshot_on', 'URG')
    kwargs.setdefault('connection_handler', 'exec')
    manhole.install(**kwargs)


class RemoteStream:
    def __init__(self, path, isatty, encoding):
        self._sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self._sock.connect(path)
        self._isatty = isatty
        self._encoding = encoding

    def isatty(self):
        return self._isatty

    def write(self, data):
        try:
            if isinstance(data, bytes):
                data = data.decode('ascii', 'replace')
            self._sock.send(data.encode(self._encoding))
        except Exception as exc:
            print(
                f'Hunter failed to send trace output {data!r} (encoding: {self._encoding!r}): {exc!r}. Stopping tracer.',
                file=sys.stderr,
            )
            hunter.stop()

    def flush(self):
        pass


@contextmanager
def manhole_bootstrap(args, activation_payload, deactivation_payload):
    activation_payload += '\nexit()\n'
    deactivation_payload += '\nexit()\n'

    activation_payload = activation_payload.encode('utf-8')
    deactivation_payload = deactivation_payload.encode('utf-8')

    with closing(connect_manhole(args.pid, args.timeout, args.signal)) as conn:
        conn.send(activation_payload)
    try:
        yield
    finally:
        with closing(connect_manhole(args.pid, args.timeout, args.signal)) as conn:
            conn.send(deactivation_payload)


@contextmanager
def gdb_bootstrap(args, activation_payload, deactivation_payload):
    print('WARNING: Using GDB may deadlock the process or create unpredictable results!')
    activation_command = [
        'gdb',
        '-p',
        str(args.pid),
        '-batch',
        '-ex',
        f'call (void)Py_AddPendingCall(PyRun_SimpleString, {json.dumps(activation_payload)})',
    ]
    deactivation_command = [
        'gdb',
        '-p',
        str(args.pid),
        '-batch',
        '-ex',
        f'call (void)Py_AddPendingCall(PyRun_SimpleString, {json.dumps(deactivation_payload)})',
    ]
    check_call(activation_command)
    try:
        yield
    finally:
        check_call(deactivation_command)


def connect_manhole(pid, timeout, signal):
    os.kill(pid, signal)

    start = time.time()
    uds_path = f'/tmp/manhole-{pid}'
    conn = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    conn.settimeout(timeout)
    while time.time() - start < timeout:
        try:
            conn.connect(uds_path)
        except OSError as exc:
            if exc.errno not in (errno.ENOENT, errno.ECONNREFUSED):
                print(
                    f'Failed to connect to {uds_path!r}: {exc!r}',
                    file=sys.stderr,
                )
        else:
            break
    else:
        print(
            f'Failed to connect to {uds_path!r}: Timeout',
            file=sys.stderr,
        )
        sys.exit(5)
    return conn


def activate(sink_path, isatty, encoding, options):
    stream = hunter._default_stream = RemoteStream(sink_path, isatty, encoding)
    try:
        stream.write('Output stream active. Starting tracer ...\n\n')
        eval(f'hunter.trace({options})')
    except Exception as exc:
        stream.write(
            'Failed to activate: {!r}. {}\n'.format(
                exc,
                f'Tracer options where: {options}.' if options else 'No tracer options.',
            )
        )
        hunter._default_stream = sys.stderr
        raise


def deactivate():
    hunter._default_stream = sys.stderr
    hunter.stop()


parser = argparse.ArgumentParser(description='Trace a process.')
parser.add_argument(
    '-p',
    '--pid',
    metavar='PID',
    type=int,
    required=True,
    help='A numerical process id.',
)
parser.add_argument('options', metavar='OPTIONS', nargs='*')
parser.add_argument(
    '-t',
    '--timeout',
    dest='timeout',
    default=1,
    type=float,
    help='Timeout to use. Default: %(default)s seconds.',
)
parser.add_argument(
    '--gdb',
    dest='gdb',
    action='store_true',
    help='Use GDB to activate tracing. WARNING: it may deadlock the process!',
)
parser.add_argument(
    '-s',
    '--signal',
    dest='signal',
    type=parse_signal,
    metavar='SIGNAL',
    default=signal.SIGURG,
    help='Send the given SIGNAL to the process before connecting. Default: %(default)s.',
)


def main():
    args = parser.parse_args()

    sink = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sink_path = f'/tmp/hunter-{os.getpid()}'
    sink.bind(sink_path)
    sink.listen(1)
    os.chmod(sink_path, 0o777)

    stdout = os.fdopen(sys.stdout.fileno(), 'wb', 0)
    encoding = getattr(sys.stdout, 'encoding', 'utf-8') or 'utf-8'
    bootstrapper = gdb_bootstrap if args.gdb else manhole_bootstrap
    payload = 'from hunter import remote; remote.activate({!r}, {!r}, {!r}, {!r})'.format(
        sink_path,
        sys.stdout.isatty(),
        encoding,
        ','.join(i.strip(',') for i in args.options),
    )
    with bootstrapper(args, payload, 'from hunter import remote; remote.deactivate()'):
        conn, _ = sink.accept()
        os.unlink(sink_path)
        pid, _, _ = get_peercred(conn)
        if pid:
            if pid != args.pid:
                raise Exception(f'Unexpected pid {pid!r} connected to output socket. Was expecting {args.pid!r}.')
        else:
            print(
                'WARNING: Failed to get pid of connected process.',
                file=sys.stderr,
            )
        data = conn.recv(1024)
        while data:
            stdout.write(data)
            data = conn.recv(1024)



# ruff: noqa: B008
import collections
import opcode
import os
import threading
from collections import defaultdict
from itertools import islice
from os import getpid
from typing import ClassVar

from . import config
from .util import BUILTIN_SYMBOLS
from .util import CALL_COLORS
from .util import CODE_COLORS
from .util import MISSING
from .util import OTHER_COLORS
from .util import builtins
from .util import frame_iterator
from .util import get_arguments
from .util import iter_symbols
from .util import safe_repr
from colorama import AnsiToWin32

try:
    from threading import get_ident
except ImportError:
    from thread import get_ident

__all__ = [
    'Action',
    'CallPrinter',
    'CodePrinter',
    'Debugger',
    'Manhole',
    'VarsPrinter',
]

BUILTIN_REPR_FUNCS = {'repr': repr, 'safe_repr': safe_repr}


class Action:
    def __call__(self, event):
        raise NotImplementedError


class LazyImportPdb:
    def __repr__(self):
        return "<class 'pdb.Pdb'>"

    __str__ = __repr__

    def __call__(self, **kwargs):
        from pdb import Pdb

        return Pdb(**kwargs)


class Debugger(Action):
    """
    An action that starts ``pdb``.
    """

    def __init__(self, klass=config.Default('klass', LazyImportPdb()), **kwargs):
        self.klass = config.resolve(klass)
        self.kwargs = kwargs

    def __eq__(self, other):
        return type(self) is type(other) and self.klass == other.klass and self.kwargs == other.kwargs

    def __str__(self):
        return f'{self.__class__.__name__}(klass={self.klass}, kwargs={self.kwargs})'

    def __repr__(self):
        return f'{self.__class__.__name__}(klass={self.klass!r}, kwargs={self.kwargs!r})'

    def __call__(self, event):
        """
        Runs a ``pdb.set_trace`` at the matching frame.
        """
        self.klass(**self.kwargs).set_trace(event.frame)


class Manhole(Action):
    def __init__(self, **options):
        self.options = options

    def __eq__(self, other):
        return type(self) is type(other) and self.options == other.options

    def __str__(self):
        return f'{self.__class__.__name__}(options={self.options})'

    def __repr__(self):
        return f'{self.__class__.__name__}(options={self.options!r})'

    def __call__(self, event):
        import manhole

        inst = manhole.install(strict=False, thread=False, **self.options)
        inst.handle_oneshot()


class ColorStreamAction(Action):
    """
    Baseclass for your custom action. Just implement your own ``__call__``.
    """

    _stream_cache: ClassVar = {}
    _stream = None
    _tty = None
    _repr_func = None

    OTHER_COLORS = OTHER_COLORS
    EVENT_COLORS = CODE_COLORS

    def __init__(
        self,
        stream=config.Default('stream', None),
        force_colors=config.Default('force_colors', False),
        force_pid=config.Default('force_pid', False),
        filename_alignment=config.Default('filename_alignment', 40),
        thread_alignment=config.Default('thread_alignment', 12),
        pid_alignment=config.Default('pid_alignment', 9),
        repr_limit=config.Default('repr_limit', 1024),
        repr_func=config.Default('repr_func', 'safe_repr'),
    ):
        self.force_colors = config.resolve(force_colors)
        self.force_pid = config.resolve(force_pid)
        stream = config.resolve(stream)
        if stream is None:
            from . import _default_stream as stream
        self.stream = stream
        self.filename_alignment = config.resolve(filename_alignment)
        self.thread_alignment = config.resolve(thread_alignment)
        self.pid_alignment = config.resolve(pid_alignment)
        self.repr_limit = config.resolve(repr_limit)
        self.repr_func = config.resolve(repr_func)
        self.seen_threads = set()
        self.seen_pid = getpid()

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self.stream == other.stream
            and self.force_colors == other.force_colors
            and self.filename_alignment == other.filename_alignment
            and self.thread_alignment == other.thread_alignment
            and self.pid_alignment == other.pid_alignment
            and self.repr_limit == other.repr_limit
            and self.repr_func == other.repr_func
        )

    def __str__(self):
        return (
            f'{self.__class__.__name__}(stream={self.stream}, force_colors={self.force_colors}, '
            f'filename_alignment={self.filename_alignment}, thread_alignment={self.thread_alignment}, '
            f'pid_alignment={self.pid_alignment} repr_limit={self.repr_limit}, '
            f'repr_func={self.repr_func})'
        )

    def __repr__(self):
        return (
            f'{self.__class__.__name__}(stream={self.stream!r}, force_colors={self.force_colors!r}, '
            f'filename_alignment={self.filename_alignment!r}, thread_alignment={self.thread_alignment!r}, '
            f'pid_alignment={self.pid_alignment!r} repr_limit={self.repr_limit!r}, '
            f'repr_func={self.repr_func!r})'
        )

    @property
    def stream(self):
        return self._stream

    @stream.setter
    def stream(self, value):
        if isinstance(value, str):
            if value in self._stream_cache:
                value = self._stream_cache[value]
            else:
                value = self._stream_cache[value] = open(value, 'a', buffering=0)

        isatty = getattr(value, 'isatty', None)
        if self.force_colors or (isatty and isatty() and os.name != 'java'):
            self._stream = AnsiToWin32(value, strip=False)
            self._tty = True
            self.event_colors = self.EVENT_COLORS
            self.other_colors = self.OTHER_COLORS
        else:
            self._tty = False
            self._stream = value
            self.event_colors = dict.fromkeys(self.EVENT_COLORS, '')
            self.other_colors = dict.fromkeys(self.OTHER_COLORS, '')

    @property
    def repr_func(self):
        return self._repr_func

    @repr_func.setter
    def repr_func(self, value):
        if callable(value):
            self._repr_func = value
        elif value in BUILTIN_REPR_FUNCS:
            self._repr_func = BUILTIN_REPR_FUNCS[value]
        else:
            raise TypeError(f'Expected a callable or either "repr" or "safe_repr" strings, not {value!r}.')

    def try_str(self, obj):
        """
        Safely call ``str(obj)``. Failures will have special colored output and output is trimmed according
        to ``self.repr_limit``.

        Only used when dumping detached events.

        Returns: string
        """
        limit = self.repr_limit
        try:
            s = str(obj)
            s = s.replace('\n', r'\n')
            if len(s) > limit:
                cutoff = limit // 2
                return '{} {CONT}[...]{RESET} {}'.format(s[:cutoff], s[-cutoff:], **self.other_colors)
            else:
                return s
        except Exception as exc:
            return '{INTERNAL-FAILURE}!!! FAILED REPR: {INTERNAL-DETAIL}{!r}{RESET}'.format(exc, **self.other_colors)

    def try_repr(self, obj):
        """
        Safely call ``self.repr_func(obj)``. Failures will have special colored output and output is trimmed according
        to ``self.repr_limit``.

        Returns: string

        """
        limit = self.repr_limit
        try:
            s = self.repr_func(obj)
            s = s.replace('\n', r'\n')
            if len(s) > limit:
                cutoff = limit // 2
                return '{} {CONT}[...]{RESET} {}'.format(s[:cutoff], s[-cutoff:], **self.other_colors)
            else:
                return s
        except Exception as exc:
            return '{INTERNAL-FAILURE}!!! FAILED REPR: {INTERNAL-DETAIL}{!r}{RESET}'.format(exc, **self.other_colors)

    def try_source(self, event, full=False):
        """
        Get a failure-colorized source for the given ``event``.

        Return: string
        """
        source = event.fullsource if full else event.source
        if source.startswith('??? NO SOURCE: '):
            return '{SOURCE-FAILURE}??? NO SOURCE: {SOURCE-DETAIL}{}'.format(source[15:], **self.other_colors)
        elif source:
            return source
        else:
            return '{SOURCE-FAILURE}??? NO SOURCE: {SOURCE-DETAIL}Source code string for {!r} is empty.'.format(
                event.filename, **self.other_colors
            )

    def filename_prefix(self, event=None):
        """
        Get an aligned and trimmed filename prefix for the given ``event``.

        Returns: string
        """
        if event:
            filename = event.filename
            if filename == '?':
                filename = '{SOURCE-FAILURE}?'.format(**self.other_colors)
            lineno = '{COLON}:{LINENO}{:<5}'.format(event.lineno, **self.other_colors)

            if len(filename) > self.filename_alignment:
                filename = f'[...]{filename[5 - self.filename_alignment :]}'
            return '{:>{}}{} '.format(filename, self.filename_alignment, lineno, **self.other_colors)
        else:
            return f'{"":>{self.filename_alignment}}       '

    def pid_prefix(self):
        """
        Get an aligned and trimmed pid prefix.
        """
        pid = getpid()
        if self.force_pid or self.seen_pid != pid:
            pid = f'[{pid}]'
            pid_align = self.pid_alignment
        else:
            pid = pid_align = ''
        return f'{pid:{pid_align}}'

    def thread_prefix(self, event):
        """
        Get an aligned and trimmed thread prefix for the given ``event``.
        """
        self.seen_threads.add(get_ident())
        if event.threading_support is False:
            threading_support = False
        elif event.threading_support:
            threading_support = True
        else:
            threading_support = len(self.seen_threads) > 1
        thread_name = threading.current_thread().name if threading_support else ''
        thread_align = self.thread_alignment if threading_support else ''
        return f'{thread_name:{thread_align}}'

    def output(self, format_str, *args, **kwargs):
        """
        Write ``format_str.format(*args, **ANSI_COLORS, **kwargs)`` to ``self.stream``.

        For ANSI coloring you can place these in the ``format_str``:

        * ``{BRIGHT}``
        * ``{DIM}``
        * ``{NORMAL}``
        * ``{RESET}``
        * ``{fore(BLACK)}``
        * ``{fore(RED)}``
        * ``{fore(GREEN)}``
        * ``{fore(YELLOW)}``
        * ``{fore(BLUE)}``
        * ``{fore(MAGENTA)}``
        * ``{fore(CYAN)}``
        * ``{fore(WHITE)}``
        * ``{fore(RESET)}``
        * ``{back(BLACK)}``
        * ``{back(RED)}``
        * ``{back(GREEN)}``
        * ``{back(YELLOW)}``
        * ``{back(BLUE)}``
        * ``{back(MAGENTA)}``
        * ``{back(CYAN)}``
        * ``{back(WHITE)}``
        * ``{back(RESET)}``

        Args:
            format_str: a PEP-3101 format string
            *args:
            **kwargs:

        Returns: string
        """
        self.stream.write(format_str.format(*args, **dict(self.other_colors, **kwargs)))

    def cleanup(self):
        pass


class CodePrinter(ColorStreamAction):
    """
    An action that just prints the code being executed.

    Args:
        stream (file-like): Stream to write to. Default: ``sys.stderr``.
        filename_alignment (int): Default size for the filename column (files are right-aligned). Default: ``40``.
        force_colors (bool): Force coloring. Default: ``False``.
        repr_limit (bool): Limit length of ``repr()`` output. Default: ``512``.
        repr_func (string or callable): Function to use instead of ``repr``.
            If string must be one of 'repr' or 'safe_repr'. Default: ``'safe_repr'``.
    """

    def __call__(self, event):
        """
        Handle event and print filename, line number and source code. If event.kind is a `return` or `exception` also
        prints values.
        """
        lines = self.try_source(event, full=True).splitlines()
        pid_prefix = self.pid_prefix()
        thread_prefix = self.thread_prefix(event)
        filename_prefix = self.filename_prefix(event)

        self.output(
            '{}{}{}{KIND}{:9} {COLOR}{}{RESET}\n',
            pid_prefix,
            thread_prefix,
            filename_prefix,
            event.kind,
            lines[0],
            COLOR=self.event_colors.get(event.kind),
        )
        if len(lines) > 1:
            empty_filename_prefix = self.filename_prefix()
            for line in lines[1:-1]:
                self.output(
                    '{}{}{}{KIND}{:9} {COLOR}{}{RESET}\n',
                    pid_prefix,
                    thread_prefix,
                    empty_filename_prefix,
                    '   |',
                    line,
                    COLOR=self.event_colors.get(event.kind),
                )
            self.output(
                '{}{}{}{KIND}{:9} {COLOR}{}{RESET}\n',
                pid_prefix,
                thread_prefix,
                empty_filename_prefix,
                '   *',
                lines[-1],
                COLOR=self.event_colors.get(event.kind),
            )

        if event.kind in ('return', 'exception'):
            self.output(
                '{}{}{}{CONT}{:9} {COLOR}{} value: {NORMAL}{}{RESET}\n',
                pid_prefix,
                thread_prefix,
                self.filename_prefix(),
                '...',
                event.kind,
                self.try_str(event.arg) if event.detached else self.try_repr(event.arg),
                COLOR=self.event_colors.get(event.kind),
            )


class CallPrinter(CodePrinter):
    """
    An action that just prints the code being executed, but unlike :obj:`hunter.CodePrinter` it indents based on
    callstack depth and it also shows ``repr()`` of function arguments.

    Args:
        stream (file-like): Stream to write to. Default: ``sys.stderr``.
        filename_alignment (int): Default size for the filename column (files are right-aligned). Default: ``40``.
        force_colors (bool): Force coloring. Default: ``False``.
        repr_limit (bool): Limit length of ``repr()`` output. Default: ``512``.
        repr_func (string or callable): Function to use instead of ``repr``.
            If string must be one of 'repr' or 'safe_repr'. Default: ``'safe_repr'``.

    .. versionadded:: 1.2.0
    """

    EVENT_COLORS = CALL_COLORS

    locals: ClassVar = defaultdict(list)

    @classmethod
    def cleanup(cls):
        cls.locals = defaultdict(list)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, event):
        """
        Handle event and print filename, line number and source code. If event.kind is a `return` or `exception` also
        prints values.
        """
        ident = event.module, event.function

        thread = threading.current_thread()
        stack = self.locals[thread.ident]

        pid_prefix = self.pid_prefix()
        thread_prefix = self.thread_prefix(event)
        filename_prefix = self.filename_prefix(event)

        if event.kind == 'call':
            if event.builtin:
                self.output(
                    '{}{}{}{KIND}{:9} {}{COLOR} >{BUILTIN} {}.{}: {}{RESET}\n',
                    pid_prefix,
                    thread_prefix,
                    filename_prefix,
                    event.kind,
                    '   ' * (len(stack) - 1),
                    event.module,
                    event.function,
                    self.try_source(event).strip(),
                    COLOR=self.event_colors.get(event.kind),
                )
            else:
                code = event.code
                stack.append(ident)
                self.output(
                    '{}{}{}{KIND}{:9} {}{COLOR}=>{NORMAL} {}({}{COLOR}{NORMAL}){RESET}\n',
                    pid_prefix,
                    thread_prefix,
                    filename_prefix,
                    event.kind,
                    '   ' * (len(stack) - 1),
                    event.function,
                    ', '.join(
                        '{VARS}{0}{VARS-NAME}{1}{VARS}={RESET}{2}'.format(
                            prefix,
                            var_display,
                            (
                                self.try_str(event.locals.get(var_lookup, MISSING))
                                if event.detached
                                else self.try_repr(event.locals.get(var_lookup, MISSING))
                            ),
                            **self.other_colors,
                        )
                        for prefix, var_lookup, var_display in get_arguments(code)
                    ),
                    COLOR=self.event_colors.get(event.kind),
                )
        elif event.kind == 'exception':
            if event.builtin:
                self.output(
                    '{}{}{}{KIND}{:9} {}{COLOR} ! {BUILTIN}{}.{}{RESET}\n',
                    pid_prefix,
                    thread_prefix,
                    filename_prefix,
                    event.kind,
                    '   ' * (len(stack) - 1),
                    event.module,
                    event.function,
                    COLOR=self.event_colors.get(event.kind),
                )
            else:
                self.output(
                    '{}{}{}{KIND}{:9} {}{COLOR} !{NORMAL} {}: {RESET}{}\n',
                    pid_prefix,
                    thread_prefix,
                    filename_prefix,
                    event.kind,
                    '   ' * (len(stack) - 1),
                    event.function,
                    self.try_str(event.arg) if event.detached else self.try_repr(event.arg),
                    COLOR=self.event_colors.get(event.kind),
                )
        elif event.kind == 'return':
            if event.builtin:
                self.output(
                    '{}{}{}{KIND}{:9} {}{COLOR} <{BUILTIN} {}.{}{RESET}\n',
                    pid_prefix,
                    thread_prefix,
                    filename_prefix,
                    event.kind,
                    '   ' * (len(stack) - 1),
                    event.module,
                    event.function,
                    COLOR=self.event_colors.get(event.kind),
                )
            else:
                self.output(
                    '{}{}{}{KIND}{:9} {}{COLOR}<={NORMAL} {}{}{RESET}{}\n',
                    pid_prefix,
                    thread_prefix,
                    filename_prefix,
                    event.kind,
                    '   ' * (len(stack) - 1),
                    event.function,
                    '' if event.builtin else ': ',
                    '' if event.builtin else (self.try_str(event.arg) if event.detached else self.try_repr(event.arg)),
                    COLOR=self.event_colors.get(event.kind),
                )
                if stack and stack[-1] == ident:
                    stack.pop()
        else:
            self.output(
                '{}{}{}{KIND}{:9} {RESET}{}{}{RESET}\n',
                pid_prefix,
                thread_prefix,
                filename_prefix,
                event.kind,
                '   ' * len(stack),
                self.try_source(event).strip(),
            )


class VarsPrinter(ColorStreamAction):
    """
    An action that prints local variables and optionally global variables visible from the current executing frame.

    Args:
        *names (strings): Names to evaluate. Expressions can be used (will only try to evaluate if all the variables are
            present on the frame.
        stream (file-like): Stream to write to. Default: ``sys.stderr``.
        filename_alignment (int): Default size for the filename column (files are right-aligned). Default: ``40``.
        force_colors (bool): Force coloring. Default: ``False``.
        repr_limit (bool): Limit length of ``repr()`` output. Default: ``512``.
        repr_func (string or callable): Function to use instead of ``repr``.
            If string must be one of 'repr' or 'safe_repr'. Default: ``'safe_repr'``.
    """

    def __init__(self, *names, **options):
        if not names:
            raise TypeError('VarsPrinter requires at least one variable name/expression.')
        self.names = {name: set(iter_symbols(name)) for name in names}
        super().__init__(**options)

    def __call__(self, event):
        """
        Handle event and print the specified variables.
        """
        first = True

        frame_symbols = set(event.locals)
        frame_symbols.update(BUILTIN_SYMBOLS)
        frame_symbols.update(event.globals)

        pid_prefix = self.pid_prefix()
        thread_prefix = self.thread_prefix(event)
        filename_prefix = self.filename_prefix(event)
        empty_filename_prefix = self.filename_prefix()

        for code, symbols in self.names.items():
            try:
                obj = eval(code, dict(vars(builtins), **event.globals), event.locals)
            except (AttributeError, KeyError):
                continue
            except Exception as exc:
                printout = '{INTERNAL-FAILURE}FAILED EVAL: {INTERNAL-DETAIL}{!r}'.format(exc, **self.other_colors)
            else:
                printout = self.try_str(obj) if event.detached else self.try_repr(obj)

            if frame_symbols >= symbols:
                if first:
                    self.output(
                        '{}{}{}{KIND}{:9} {VARS}[{VARS-NAME}{} {VARS}=> {RESET}{}{VARS}]{RESET}\n',
                        pid_prefix,
                        thread_prefix,
                        filename_prefix,
                        event.kind,
                        code,
                        printout,
                    )
                    first = False
                else:
                    self.output(
                        '{}{}{}{CONT}...       {VARS}[{VARS-NAME}{} {VARS}=> {RESET}{}{VARS}]{RESET}\n',
                        pid_prefix,
                        thread_prefix,
                        empty_filename_prefix,
                        code,
                        printout,
                    )


class VarsSnooper(ColorStreamAction):
    """
    A PySnooper-inspired action, similar to :class:`~hunter.actions.VarsPrinter`, but only show variable changes.

    .. warning:: Should be considered experimental. Use judiciously.

        * It stores reprs for all seen variables, therefore it can use lots of memory.
        * Will leak memory if you filter the return events (eg: ``~Q(kind="return")``).
        * Not thoroughly tested. May misbehave on code with closures/nonlocal variables.

    Args:
        stream (file-like): Stream to write to. Default: ``sys.stderr``.
        filename_alignment (int): Default size for the filename column (files are right-aligned). Default: ``40``.
        force_colors (bool): Force coloring. Default: ``False``.
        repr_limit (bool): Limit length of ``repr()`` output. Default: ``512``.
        repr_func (string or callable): Function to use instead of ``repr``.
            If string must be one of 'repr' or 'safe_repr'. Default: ``'safe_repr'``.
    """

    def __init__(self, **options):
        super().__init__(**options)
        self.stored_reprs = defaultdict(dict)

    def __call__(self, event):
        """
        Handle event and print the specified variables.
        """
        first = True

        pid_prefix = self.pid_prefix()
        thread_prefix = self.thread_prefix(event)
        filename_prefix = self.filename_prefix(event)
        empty_filename_prefix = self.filename_prefix()

        current_reprs = {name: self.try_str(value) if event.detached else self.try_repr(value) for name, value in event.locals.items()}
        scope_key = event.code or event.function
        scope = self.stored_reprs[scope_key]
        for name, current_repr in sorted(current_reprs.items()):
            previous_repr = scope.get(name)
            if previous_repr is None:
                scope[name] = current_repr
                if first:
                    self.output(
                        '{}{}{}{KIND}{:9} {VARS}[{VARS-NAME}{} {VARS}:= {RESET}{}{VARS}]{RESET}\n',
                        pid_prefix,
                        thread_prefix,
                        filename_prefix,
                        event.kind,
                        name,
                        current_repr,
                    )
                    first = False
                else:
                    self.output(
                        '{}{}{}{CONT}{:9} {VARS}[{VARS-NAME}{} {VARS}:= {RESET}{}{VARS}]{RESET}\n',
                        pid_prefix,
                        thread_prefix,
                        empty_filename_prefix,
                        '...',
                        name,
                        current_repr,
                    )
            elif previous_repr != current_repr:
                scope[name] = current_repr
                if first:
                    self.output(
                        '{}{}{}{KIND}{:9} {VARS}[{VARS-NAME}{} {VARS}: {RESET}{}{VARS} => {RESET}{}{VARS}]{RESET}\n',
                        pid_prefix,
                        thread_prefix,
                        filename_prefix,
                        event.kind,
                        name,
                        previous_repr,
                        current_repr,
                    )
                    first = False
                else:
                    self.output(
                        '{}{}{}{CONT}{:9} {VARS}[{VARS-NAME}{} {VARS}: {RESET}{}{VARS} => {RESET}{}{VARS}]{RESET}\n',
                        pid_prefix,
                        thread_prefix,
                        empty_filename_prefix,
                        '...',
                        name,
                        previous_repr,
                        current_repr,
                    )
        if event.kind == 'return':
            if scope_key in self.stored_reprs:
                del self.stored_reprs[scope_key]


RETURN_OPCODES = (opcode.opmap['RETURN_VALUE'],)
if 'RETURN_CONST' in opcode.opmap:
    RETURN_OPCODES += (opcode.opmap['RETURN_CONST'],)


class ErrorSnooper(CodePrinter):
    """
    An action that prints events around silenced exceptions. Note that it inherits the output of :class:`~hunter.actions.CodePrinter` so no
    fancy call indentation.

    .. warning::

        Should be considered experimental. May show lots of false positives especially if you're tracing lots of clumsy code like:

        .. sourcecode:: python

            try:
                stuff = something[key]
            except KeyError:
                stuff = "default"

    Args:
        max_backlog (int): Maximum number of events to record and display before the silenced exception is raised.
            Set to 0 to disable and get a speed boost. Default: ``10``.
        max_events (int): Maximum number of events to record and display for each detected silenced exception. Default: ``50``.
        max_depth (int): Increase if you want to drill into subsequent calls after an exception is raised. If you increase this you might
            want to also increase ``max_events`` since subsequent calls may have so many events you won't get to see the return event.
            Default: ``0`` (doesn't drill into any calls).

        stream (file-like): Stream to write to. Default: ``sys.stderr``.
        filename_alignment (int): Default size for the filename column (files are right-aligned). Default: ``40``.
        force_colors (bool): Force coloring. Default: ``False``.
        repr_limit (bool): Limit length of ``repr()`` output. Default: ``512``.
        repr_func (string or callable): Function to use instead of ``repr``.
            If string must be one of 'repr' or 'safe_repr'. Default: ``'safe_repr'``.

    .. versionadded:: 3.1.0
    """

    def __init__(self, *args, **kwargs):
        self.max_backlog = kwargs.pop('max_backlog', 10)
        self.backlog = collections.deque(maxlen=self.max_backlog)
        self.max_events = kwargs.pop('max_events', 50)
        self.max_depth = kwargs.pop('max_depth', 0)
        self.origin = None
        self.events = None
        super().__init__(*args, **kwargs)

    def __call__(self, event):
        if self.max_backlog:
            detached_event = event.detach(self.try_repr)
            self.backlog.append(detached_event)
        else:
            detached_event = None

        if event.kind == 'exception':  # something interesting happened ;)
            if detached_event is None:
                detached_event = event.detach(self.try_repr)
            if self.origin:
                self.events.append(detached_event)
                if self.origin.depth > event.depth:
                    self.origin = detached_event
            else:
                self.events = list(self.backlog)
                self.origin = detached_event
        elif self.origin:
            if event.kind == 'return':
                if detached_event is None:
                    detached_event = event.detach(self.try_repr)
                self.events.append(detached_event)
                if event.depth == self.origin.depth - 1:  # stop if the same function returned (depth is -1)
                    if event.instruction in RETURN_OPCODES:
                        self.dump_events()
                        self.output(
                            '{BRIGHT}{fore(BLACK)}{} function exit{RESET}\n',
                            '-' * 46,
                        )
                    self.origin = None
                    self.events = None
            elif event.depth > self.origin.depth + self.max_depth:  # too many details
                return
            elif len(self.events) > self.max_events:
                self.dump_events()
                self.output('{BRIGHT}{fore(BLACK)}{} too many lines{RESET}\n', '-' * 46)
            else:
                if detached_event is None:
                    detached_event = event.detach(self.try_repr)
                self.events.append(detached_event)

    def dump_events(self):
        self.output(
            '{BRIGHT}{fore(BLUE)}{} tracing {fore(YELLOW)}{}{fore(BLUE)} on {fore(RED)}{}{RESET}\n',
            '>' * 46,
            self.origin.function,
            self.origin.arg,
        )
        for event in self.events:
            super().__call__(event)
        self.origin = None
        self.events = None


class StackPrinter(ColorStreamAction):
    """
    An action that prints a one-line stacktrace.

    Args:
        depth (int): The maximum number of frames to show.
        limit (int): The maximum number of components to show in path. Eg: ``limit=2`` means it will show 1 parent: ``foo/bar.py``.
        stream (file-like): Stream to write to. Default: ``sys.stderr``.
        filename_alignment (int): Default size for the filename column (files are right-aligned). Default: ``40``.
        force_colors (bool): Force coloring. Default: ``False``.
        repr_limit (bool): Limit length of ``repr()`` output. Default: ``512``.
        repr_func (string or callable): Function to use instead of ``repr``.
            If string must be one of 'repr' or 'safe_repr'. Default: ``'safe_repr'``.
    """

    def __init__(self, depth=15, limit=2, **options):
        self.limit = limit
        self.depth = depth
        super().__init__(**options)

    def __call__(self, event):
        """
        Handle event and print the stack.
        """
        pid_prefix = self.pid_prefix()
        thread_prefix = self.thread_prefix(event)
        filename_prefix = self.filename_prefix(event).rstrip()
        sep = os.path.sep

        if event.frame and event.frame.f_back:
            template = '{{}}{{}}{{}}{{CONT}}:{{BRIGHT}}{{fore(BLUE)}}{} {{KIND}}<={{RESET}} {}'.format(
                event.function,
                ' {KIND}<={RESET} '.join(
                    '{}{{CONT}}:{{RESET}}{}{{CONT}}:{{BRIGHT}}{{fore(BLUE)}}{}'.format(
                        '/'.join(frame.f_code.co_filename.split(sep)[-self.limit :]),
                        frame.f_lineno,
                        frame.f_code.co_name,
                    )
                    for frame in islice(frame_iterator(event.frame.f_back), self.depth)
                ),
            )
        else:
            template = f'{{}}{{}}{{}}{{CONT}}:{{BRIGHT}}{{fore(BLUE)}}{event.function} {{KIND}}<= {{BRIGHT}}{{fore(YELLOW)}}no frames available {{NORMAL}}(detached={event.detached})'
        template += '{RESET}\n'
        self.output(
            template,
            pid_prefix,
            thread_prefix,
            filename_prefix,
        )



import collections
import inspect
import re
from itertools import chain

from .actions import Action
from .actions import ColorStreamAction
from .event import Event

__all__ = (
    'And',
    'From',
    'Not',
    'Or',
    'Query',
    'When',
)

ALLOWED_KEYS = tuple(sorted(i for i in Event.__dict__.keys() if not i.startswith('_') and i not in ('tracer', 'thread', 'frame')))
ALLOWED_OPERATORS = (
    'startswith',
    'endswith',
    'in',
    'contains',
    'regex',
    'sw',
    'ew',
    'has',
    'rx',
    'gt',
    'gte',
    'lt',
    'lte',
)


class BasePredicate:
    pass


class Query:
    """
    Event-filtering predicate.

    See :class:`hunter.event.Event` for details about the fields that can be filtered on.

    Args:
        query: criteria to match on.

            Accepted arguments:
            ``arg``,
            ``builtin``,
            ``calls``,
            ``code``,
            ``depth``,
            ``filename``,
            ``frame``,
            ``fullsource``,
            ``function``,
            ``globals``,
            ``kind``,
            ``lineno``,
            ``locals``,
            ``module``,
            ``source``,
            ``stdlib``,
            ``threadid``,
            ``threadname``.
    """

    def __init__(self, **query):
        query_eq = {}
        query_startswith = {}
        query_endswith = {}
        query_in = {}
        query_contains = {}
        query_regex = {}
        query_lt = {}
        query_lte = {}
        query_gt = {}
        query_gte = {}

        for key, value in query.items():
            parts = [p for p in key.split('_') if p]
            count = len(parts)
            if count > 2:
                raise TypeError(
                    f'Unexpected argument {key!r}. Must be one of {ALLOWED_KEYS} with optional operators like: {ALLOWED_OPERATORS}'
                )
            elif count == 2:
                prefix, operator = parts
                if operator in ('startswith', 'sw'):
                    if not isinstance(value, str):
                        if not isinstance(value, (list, set, tuple)):
                            raise ValueError(f'Value {value!r} for {key!r} is invalid. Must be a string, list, tuple or set.')
                        value = tuple(value)
                    mapping = query_startswith
                elif operator in ('endswith', 'ew'):
                    if not isinstance(value, str):
                        if not isinstance(value, (list, set, tuple)):
                            raise ValueError(f'Value {value!r} for {key!r} is invalid. Must be a string, list, tuple or set.')
                        value = tuple(value)
                    mapping = query_endswith
                elif operator == 'in':
                    mapping = query_in
                elif operator in ('contains', 'has'):
                    mapping = query_contains
                elif operator in ('regex', 'rx'):
                    value = re.compile(value)
                    mapping = query_regex
                elif operator == 'lt':
                    mapping = query_lt
                elif operator == 'lte':
                    mapping = query_lte
                elif operator == 'gt':
                    mapping = query_gt
                elif operator == 'gte':
                    mapping = query_gte
                else:
                    raise TypeError(f'Unexpected operator {operator!r}. Must be one of {ALLOWED_OPERATORS}.')
            else:
                mapping = query_eq
                prefix = key

            if prefix not in ALLOWED_KEYS:
                raise TypeError(f'Unexpected argument {key!r}. Must be one of {ALLOWED_KEYS}.')

            mapping[prefix] = value

        self.query_eq = tuple(sorted(query_eq.items()))
        self.query_startswith = tuple(sorted(query_startswith.items()))
        self.query_endswith = tuple(sorted(query_endswith.items()))
        self.query_in = tuple(sorted(query_in.items()))
        self.query_contains = tuple(sorted(query_contains.items()))
        self.query_regex = tuple(sorted(query_regex.items()))
        self.query_lt = tuple(sorted(query_lt.items()))
        self.query_lte = tuple(sorted(query_lte.items()))
        self.query_gt = tuple(sorted(query_gt.items()))
        self.query_gte = tuple(sorted(query_gte.items()))

    def __str__(self):
        return 'Query(%s)' % (  # noqa: UP031
            ', '.join(
                ', '.join(f'{key}{kind}={value!r}' for key, value in mapping)
                for kind, mapping in [
                    ('', self.query_eq),
                    ('_in', self.query_in),
                    ('_contains', self.query_contains),
                    ('_startswith', self.query_startswith),
                    ('_endswith', self.query_endswith),
                    ('_regex', self.query_regex),
                    ('_lt', self.query_lt),
                    ('_lte', self.query_lte),
                    ('_gt', self.query_gt),
                    ('_gte', self.query_gte),
                ]
                if mapping
            )
        )

    def __repr__(self):
        return '<hunter.predicates.Query: %s>' % ' '.join(  # noqa: UP031
            fmt % (mapping,)
            for fmt, mapping in [
                ('query_eq=%r', self.query_eq),
                ('query_in=%r', self.query_in),
                ('query_contains=%r', self.query_contains),
                ('query_startswith=%r', self.query_startswith),
                ('query_endswith=%r', self.query_endswith),
                ('query_regex=%r', self.query_regex),
                ('query_lt=%r', self.query_lt),
                ('query_lte=%r', self.query_lte),
                ('query_gt=%r', self.query_gt),
                ('query_gte=%r', self.query_gte),
            ]
            if mapping
        )

    def __eq__(self, other):
        return (
            isinstance(other, Query)
            and self.query_eq == other.query_eq
            and self.query_in == other.query_in
            and self.query_contains == other.query_contains
            and self.query_startswith == other.query_startswith
            and self.query_endswith == other.query_endswith
            and self.query_regex == other.query_regex
            and self.query_lt == other.query_lt
            and self.query_lte == other.query_lte
            and self.query_gt == other.query_gt
            and self.query_gte == other.query_gte
        )

    def __call__(self, event):
        """
        Handles event. Returns True if all criteria matched.
        """
        for key, value in self.query_eq:
            evalue = event[key]
            if evalue != value:
                return False
        for key, value in self.query_in:
            evalue = event[key]
            if evalue not in value:
                return False
        for key, value in self.query_contains:
            evalue = event[key]
            if value not in evalue:
                return False
        for key, value in self.query_startswith:
            evalue = event[key]
            if not evalue.startswith(value):
                return False
        for key, value in self.query_endswith:
            evalue = event[key]
            if not evalue.endswith(value):
                return False
        for key, value in self.query_regex:
            evalue = event[key]
            if not value.match(evalue):
                return False
        for key, value in self.query_gt:
            evalue = event[key]
            if not evalue > value:
                return False
        for key, value in self.query_gte:
            evalue = event[key]
            if not evalue >= value:
                return False
        for key, value in self.query_lt:
            evalue = event[key]
            if not evalue < value:
                return False
        for key, value in self.query_lte:
            evalue = event[key]
            if not evalue <= value:
                return False

        return True

    def __or__(self, other):
        """
        Convenience API so you can do ``Query(...) | Query(...)``. It converts that to ``Or(Query(...), Query(...))``.
        """
        return Or(self, other)

    def __and__(self, other):
        """
        Convenience API so you can do ``Query(...) & Query(...)``. It converts that to ``And(Query(...), Query(...))``.
        """
        return And(self, other)

    def __invert__(self):
        """
        Convenience API so you can do ``~Query(...)``. It converts that to ``Not(Query(...))``.
        """
        return Not(self)

    def __ror__(self, other):
        """
        Convenience API so you can do ``other | Query(...)``. It converts that to ``Or(other, Query(...))``.
        """
        return Or(other, self)

    def __rand__(self, other):
        """
        Convenience API so you can do ``other & Query(...)``. It converts that to ``And(other, Query(...))``.
        """
        return And(other, self)


class When:
    """
    Conditional predicate. Runs ``actions`` when ``condition(event)`` is ``True``.

    Actions take a single ``event`` argument.
    """

    def __init__(self, condition, *actions):
        if not actions:
            raise TypeError('Must give at least one action.')
        self.condition = condition
        self.actions = tuple(action() if inspect.isclass(action) and issubclass(action, Action) else action for action in actions)

    def __str__(self):
        return 'When({}, {})'.format(
            self.condition,
            ', '.join(repr(p) for p in self.actions),
        )

    def __repr__(self):
        return f'<hunter.predicates.When: condition={self.condition!r}, actions={self.actions!r}>'

    def __eq__(self, other):
        return isinstance(other, When) and self.condition == other.condition and self.actions == other.actions

    def __call__(self, event):
        """
        Handles the event.
        """
        if self.condition(event):
            for action in self.actions:
                action(event)
            return True
        else:
            return False

    def __or__(self, other):
        """
        Convenience API so you can do ``When(...) | other``. It converts that to ``Or(When(...), other)``.
        """
        return Or(self, other)

    def __and__(self, other):
        """
        Convenience API so you can do ``When(...) & other``. It converts that to ``And(When(...), other)``.
        """
        return And(self, other)

    def __invert__(self):
        """
        Convenience API so you can do ``~When(...)``. It converts that to ``Not(When(...))``.
        """
        return Not(self)

    def __ror__(self, other):
        """
        Convenience API so you can do ``other | When(...)``. It converts that to ``Or(other, When(...))``.
        """
        return Or(other, self)

    def __rand__(self, other):
        """
        Convenience API so you can do ``other & When(...)``. It converts that to ``And(other, When(...))``.
        """
        return And(other, self)


class From:
    """
    From-point filtering mechanism. Switches on to running the predicate after condition matches, and switches off when
    the depth goes lower than the initial level.

    After ``condition(event)`` returns ``True`` the ``event.depth`` will be saved and calling this object with an
    ``event`` will return ``predicate(event)`` until ``event.depth - watermark`` is equal to the depth that was saved.

    Args:
        condition (callable): Optional :class:`~hunter.predicates.Query` object or a callable that returns True/False.
        predicate (callable): Optional :class:`~hunter.predicates.Query` object or a callable that returns True/False to
            run after ``condition`` first returns ``True``. Note that this predicate will be called with a event-copy that has adjusted
            :attr:`~hunter.event.Event.depth` and :attr:`~hunter.event.Event.calls` to the initial point where the ``condition`` matched.
            In other words they will be relative.
        watermark (int): Depth difference to switch off and wait again on ``condition``.

    See Also:
         :class:`hunter.predicates.Backlog`
    """

    def __init__(self, condition, predicate=None, watermark=0):
        self.condition = condition
        self.predicate = predicate
        self.watermark = watermark
        self._origin_depth = None
        self._origin_calls = None

    def __str__(self):
        return f'From({self.condition}, {self.predicate}, watermark={self.watermark})'

    def __repr__(self):
        return f'<hunter.predicates.From: condition={self.condition!r}, predicate={self.predicate!r}, watermark={self.watermark!r}>'

    def __eq__(self, other):
        return isinstance(other, From) and self.condition == other.condition and self.predicate == other.predicate

    def __call__(self, event):
        """
        Handles the event.
        """
        if self._origin_depth is None:
            if self.condition(event):
                self._origin_depth = event.depth
                self._origin_calls = event.calls
                delta_depth = delta_calls = 0
            else:
                return False
        else:
            delta_depth = event.depth - self._origin_depth
            delta_calls = event.calls - self._origin_calls
            if delta_depth < self.watermark:
                self._origin_depth = None
                return False
        if self.predicate is None:
            return True
        else:
            relative_event = event.clone()
            relative_event.depth = delta_depth
            relative_event.calls = delta_calls
            return self.predicate(relative_event)

    def __or__(self, other):
        """
        Convenience API so you can do ``From(...) | other``. It converts that to ``Or(From(...), other)``.
        """
        return Or(self, other)

    def __and__(self, other):
        """
        Convenience API so you can do ``From(...) & other``. It converts that to ``And(From(...), other))``.
        """
        return And(self, other)

    def __invert__(self):
        """
        Convenience API so you can do ``~From(...)``. It converts that to ``Not(From(...))``.
        """
        return Not(self)

    def __ror__(self, other):
        """
        Convenience API so you can do ``other | From(...)``. It converts that to ``Or(other, From(...))``.
        """
        return Or(other, self)

    def __rand__(self, other):
        """
        Convenience API so you can do ``other & From(...)``. It converts that to ``And(other, From(...))``.
        """
        return And(other, self)


class And:
    """
    Logical conjunction. Returns ``False`` at the first sub-predicate that returns ``False``, otherwise returns ``True``.
    """

    def __init__(self, *predicates):
        self.predicates = predicates

    def __str__(self):
        return f'And({", ".join(str(p) for p in self.predicates)})'

    def __repr__(self):
        return f'<hunter.predicates.And: predicates={self.predicates!r}>'

    def __eq__(self, other):
        return isinstance(other, And) and self.predicates == other.predicates

    def __call__(self, event):
        """
        Handles the event.
        """
        for predicate in self.predicates:
            if not predicate(event):
                return False
        else:
            return True

    def __or__(self, other):
        """
        Convenience API so you can do ``And(...) | other``. It converts that to ``Or(And(...), other)``.
        """
        return Or(self, other)

    def __and__(self, other):
        """
        Convenience API so you can do ``And(...) & other``. It converts that to ``And(..., other)``.
        """
        return And(
            *chain(
                self.predicates,
                other.predicates if isinstance(other, And) else (other,),
            )
        )

    def __invert__(self):
        """
        Convenience API so you can do ``~And(...)``. It converts that to ``Not(And(...))``.
        """
        return Not(self)

    def __ror__(self, other):
        """
        Convenience API so you can do ``other | And(...)``. It converts that to ``Or(other, And(...))``.
        """
        return Or(other, self)

    def __rand__(self, other):
        """
        Convenience API so you can do ``other & And(...)``. It converts that to ``And(other, And(...))``.
        """
        return And(other, *self.predicates)


class Or:
    """
    Logical disjunction. Returns ``True`` after the first sub-predicate that returns ``True``.
    """

    def __init__(self, *predicates):
        self.predicates = predicates

    def __str__(self):
        return f'Or({", ".join(str(p) for p in self.predicates)})'

    def __repr__(self):
        return f'<hunter.predicates.Or: predicates={self.predicates!r}>'

    def __eq__(self, other):
        return isinstance(other, Or) and self.predicates == other.predicates

    def __call__(self, event):
        """
        Handles the event.
        """
        for predicate in self.predicates:
            if predicate(event):
                return True
        else:
            return False

    def __or__(self, other):
        """
        Convenience API so you can do ``Or(...) | other``. It converts that to ``Or(..., other)``.
        """
        return Or(
            *chain(
                self.predicates,
                other.predicates if isinstance(other, Or) else (other,),
            )
        )

    def __and__(self, other):
        """
        Convenience API so you can do ``Or(...) & other``. It converts that to ``And(Or(...), other)``.
        """
        return And(self, other)

    def __invert__(self):
        """
        Convenience API so you can do ``~Or(...)``. It converts that to ``Not(Or(...))``.
        """
        return Not(self)

    def __ror__(self, other):
        """
        Convenience API so you can do ``other | Or(...)``. It converts that to ``Or(other, Or(...))``.
        """
        return Or(other, *self.predicates)

    def __rand__(self, other):
        """
        Convenience API so you can do ``other & Or(...)``. It converts that to ``And(other, Or(...))``.
        """
        return And(other, self)


class Not:
    """
    Logical complement (negation). Simply returns ``not predicate(event)``.
    """

    def __init__(self, predicate):
        self.predicate = predicate

    def __str__(self):
        return f'Not({self.predicate})'

    def __repr__(self):
        return f'<hunter.predicates.Not: predicate={self.predicate!r}>'

    def __eq__(self, other):
        return isinstance(other, Not) and self.predicate == other.predicate

    def __call__(self, event):
        """
        Handles the event.
        """
        return not self.predicate(event)

    def __or__(self, other):
        """
        Convenience API so you can do ``Not(...) | other``. It converts that to ``Or(Not(...), other)``.

        Note that ``Not(...) | Not(...)`` converts to ``Not(And(..., ...))``.
        """
        if isinstance(other, Not):
            return Not(And(self.predicate, other.predicate))
        else:
            return Or(self, other)

    def __and__(self, other):
        """
        Convenience API so you can do ``Not(...) & other``. It converts that to ``And(Not(...), other)``.

        Note that ``Not(...) & Not(...)`` converts to ``Not(Or(..., ...))``.
        """
        if isinstance(other, Not):
            return Not(Or(self.predicate, other.predicate))
        else:
            return And(self, other)

    def __invert__(self):
        """
        Convenience API so you can do ``~Not(...)``. It converts that to ``...``.
        """
        return self.predicate

    def __ror__(self, other):
        """
        Convenience API so you can do ``other | Not(...)``. It converts that to ``Or(other, Not(...))``.
        """
        return Or(other, self)

    def __rand__(self, other):
        """
        Convenience API so you can do ``other & Not(...)``. It converts that to ``And(other, Not(...))``.
        """
        return And(other, self)


class Backlog:
    """
    Until-point buffering mechanism. It will buffer detached events up to the given ``size`` and display them
    using the given ``action`` when ``condition`` returns True.

    This is a complement to :class:`~hunter.predicates.From` - essentially working the other way. While :class:`~hunter.predicates.From`
    shows events after something interesting occurred the Backlog will show events prior to something interesting occurring.

    If the depth delta from the first event in the backlog and the event that matched the condition is less
    than the given ``stack`` then it will create fake call events to be passed to the action before the events
    from the backlog are passed in.

    Using a ``filter`` or pre-filtering is recommended to reduce storage work and improve tracing speed. Pre-filtering means that you use
    Backlog inside a :class:`~hunter.When` or `:class:`~hunter.And` - effectively reducing the number of Events that get to the Backlog.

    Args:
        condition (callable): Optional :class:`~hunter.predicates.Query` object or a callable that returns True/False.
        size (int): Number of events that the backlog stores. Effectively this is the ``maxlen`` for the internal deque.
        stack (int): Stack size to fill. Setting this to ``0`` disables creating fake call events.
        vars (bool): Makes global/local variables available in the stored events.
            This is an expensive option - it will use ``action.try_repr`` on all the variables.
        strip (bool): If this option is set then the backlog will be cleared every time an event matching the ``condition`` is found.
            Disabling this may show more context every time an event matching the ``condition`` is found but said context may also be
            duplicated across multiple matches.
        action (ColorStreamAction): A ColorStreamAction to display the stored events when an event matching the ``condition`` is found.
        filter (callable): Optional :class:`~hunter.predicates.Query` object or a callable that returns True/False to filter the stored
            events with.

    See Also:
         :class:`hunter.predicates.From`
    """

    def __init__(
        self,
        condition,
        size=100,
        stack=10,
        vars=False,
        strip=True,
        action=None,
        filter=None,
    ):
        self.action = action() if inspect.isclass(action) and issubclass(action, Action) else action
        if not isinstance(self.action, ColorStreamAction):
            raise TypeError(f'Action {self.action!r} must be a ColorStreamAction.')
        self.condition = condition
        self.queue = collections.deque(maxlen=size)
        self.size = size
        self.stack = stack
        self.strip = strip
        self.vars = vars
        self._try_repr = self.action.try_repr if self.vars else None
        self._filter = filter

    def __call__(self, event):
        """
        Handles the event.
        """
        result = self.condition(event)
        if result:
            if self.queue:
                self.action.cleanup()

                first_event = self.queue[0]
                first_depth = first_event.depth
                backlog_call_depth = event.depth - first_depth
                first_is_call = first_event.kind == 'call'  # note that True is 1, thus the following math is valid
                missing_depth = min(
                    first_depth,
                    max(0, self.stack - backlog_call_depth + first_is_call),
                )
                if missing_depth:
                    if first_is_call and first_event.frame is not None:
                        first_frame = first_event.frame.f_back
                    else:
                        first_frame = first_event.frame
                    if first_frame is not None:
                        stack_events = collections.deque()  # a new deque because self.queue is limited, we can't add while it's full
                        frame = first_frame
                        depth_delta = 0
                        while frame and depth_delta < missing_depth:
                            stack_event = Event(
                                frame=frame,
                                kind='call',
                                arg=None,
                                threading_support=event.threading_support,
                                depth=first_depth - depth_delta - 1,
                                calls=-1,
                            )
                            if not self.vars:
                                # noinspection PyPropertyAccess
                                stack_event.locals = {}
                                stack_event.globals = {}
                                stack_event.detached = True
                            stack_events.appendleft(stack_event)
                            frame = frame.f_back
                            depth_delta += 1
                        for stack_event in stack_events:
                            if self._filter is None or self._filter(stack_event):
                                self.action(stack_event)
                for backlog_event in self.queue:
                    if self._filter is None:
                        self.action(backlog_event)
                    elif self._filter(backlog_event):
                        self.action(backlog_event)
                self.queue.clear()
        else:
            if self.strip and event.depth < 1:
                # Looks like we're back to depth 0 for some reason.
                # Delete everything because we don't want to see what is likely just a long stream of useless returns.
                self.queue.clear()
            if self._filter is None or self._filter(event):
                detached_event = event.detach(self._try_repr)
                detached_event.frame = event.frame
                self.queue.append(detached_event)

        return result

    def __str__(self):
        return f'Backlog({self.condition}, size={self.size}, stack={self.stack}, vars={self.vars}, action={self.action}, filter={self._filter})'

    def __repr__(self):
        return f'<hunter.predicates.Backlog: condition={self.condition!r}, size={self.size!r}, stack={self.stack!r}, vars={self.vars!r}, action={self.action!r}, filter={self._filter!r}>'

    def __eq__(self, other):
        return (
            isinstance(other, Backlog)
            and self.condition == other.condition
            and self.size == other.size
            and self.stack == other.stack
            and self.vars == other.vars
            and self.action == other.action
        )

    def __or__(self, other):
        """
        Convenience API so you can do ``Backlog(...) | other``. It converts that to ``Or(Backlog(...), other)``.
        """
        return Or(self, other)

    def __and__(self, other):
        """
        Convenience API so you can do ``Backlog(...) & other``. It converts that to ``And(Backlog(...), other))``.
        """
        return And(self, other)

    def __invert__(self):
        """
        Convenience API so you can do ``~Backlog(...)``. It converts that to ``Not(Backlog(...))``.
        """
        return Backlog(
            Not(self.condition),
            size=self.size,
            stack=self.stack,
            vars=self.vars,
            action=self.action,
            filter=self._filter,
        )

    def __ror__(self, other):
        """
        Convenience API so you can do ``other | Backlog(...)``. It converts that to ``Or(other, Backlog(...))``.
        """
        return Or(other, self)

    def __rand__(self, other):
        """
        Convenience API so you can do ``other & Backlog(...)``. It converts that to ``And(other, Backlog(...))``.
        """
        return And(other, self)

    def filter(self, *predicates, **kwargs):
        """
        Returns another Backlog instance with extra output filtering. If the current instance already
        have filters they will be merged by using an :class:`~hunter.predicates.And` predicate.

        Args:
            *predicates (callables): Callables that returns True/False or :class:`~hunter.predicates.Query` objects.
            **kwargs: Arguments that may be passed to :class:`~hunter.predicates.Query`.

        Returns: A new :class:`~hunter.predicates.Backlog` instance.
        """
        from hunter import _merge

        if self._filter is not None:
            predicates = (self._filter, *predicates)

        return Backlog(
            self.condition,
            size=self.size,
            stack=self.stack,
            vars=self.vars,
            action=self.action,
            filter=_merge(*predicates, **kwargs),
        )



class Default:
    def __init__(self, key, fallback_value):
        self.key = key
        self.fallback_value = fallback_value

    def resolve(self):
        from . import _default_config

        return _default_config.get(self.key, self.fallback_value)

    def __str__(self):
        return str(self.fallback_value)

    def __repr__(self):
        return repr(self.fallback_value)


def resolve(value):
    if isinstance(value, Default):
        return value.resolve()
    else:
        return value



import sys
import threading
import traceback

import hunter

from .event import Event

__all__ = ('Tracer',)


class Tracer:
    """
    Tracer object.

    Args:
        threading_support (bool): Hooks the tracer into ``threading.settrace`` as well if True.
    """

    def __init__(self, threading_support=None, profiling_mode=False):
        self._handler = None
        self._previous = None
        self._threading_previous = None

        #: True if threading support was enabled. Should be considered read-only.
        #:
        #: :type: bool
        self.threading_support = threading_support

        #: True if profiling mode was enabled. Should be considered read-only.
        #:
        #: :type: bool
        self.profiling_mode = profiling_mode

        #: Tracing depth (increases on calls, decreases on returns)
        #:
        #: :type: int
        self.depth = 0

        #: A counter for total number of 'call' frames that this Tracer went through.
        #:
        #: :type: int
        self.calls = 0

    @property
    def handler(self):
        """
        The current predicate. Set via :func:`hunter.Tracer.trace`.
        """
        return self._handler

    @property
    def previous(self):
        """
        The previous tracer, if any (whatever ``sys.gettrace()`` returned prior to :func:`hunter.Tracer.trace`).
        """
        return self._previous

    def __repr__(self):
        return '<hunter.tracer.Tracer at 0x{:x}: threading_support={}, {}{}{}{}>'.format(
            id(self),
            self.threading_support,
            '<stopped>' if self._handler is None else 'handler=',
            '' if self._handler is None else repr(self._handler),
            '' if self._previous is None else ', previous=',
            '' if self._previous is None else repr(self._previous),
        )

    def __call__(self, frame, kind, arg):
        """
        The settrace function.

        .. note::

            This always returns self (drills down) - as opposed to only drilling down when ``predicate(event)`` is True
            because it might match further inside.
        """
        if self._handler is not None:
            if kind == 'return' and self.depth > 0:
                self.depth -= 1
            event = Event(frame, kind, arg, self.depth, self.calls, self.threading_support)
            try:
                self._handler(event)
            except Exception as exc:
                traceback.print_exc(file=hunter._default_stream)
                hunter._default_stream.write(f'Disabling tracer because handler {self._handler!r} failed ({exc!r}) at {event!r}.\n\n')
                self.stop()
                return
            if kind == 'call':
                self.depth += 1
                self.calls += 1

            return self

    def trace(self, predicate):
        """
        Starts tracing with the given callable.

        Args:
            predicate (callable that accepts a single :obj:`~hunter.event.Event` argument):
        Return:
            self
        """
        self._handler = predicate
        if self.profiling_mode:
            if self.threading_support is None or self.threading_support:
                self._threading_previous = getattr(threading, '_profile_hook', None)
                threading.setprofile(self)
            self._previous = sys.getprofile()
            sys.setprofile(self)
        else:
            if self.threading_support is None or self.threading_support:
                self._threading_previous = getattr(threading, '_trace_hook', None)
                threading.settrace(self)
            self._previous = sys.gettrace()
            sys.settrace(self)
        return self

    def stop(self):
        """
        Stop tracing. Reinstalls the :attr:`~hunter.tracer.Tracer.previous` tracer.
        """
        if self._handler is not None:
            if self.profiling_mode:
                sys.setprofile(self._previous)
                self._handler = self._previous = None
                if self.threading_support is None or self.threading_support:
                    threading.setprofile(self._threading_previous)
                    self._threading_previous = None
            else:
                sys.settrace(self._previous)
                self._handler = self._previous = None
                if self.threading_support is None or self.threading_support:
                    threading.settrace(self._threading_previous)
                    self._threading_previous = None

    def __enter__(self):
        """
        Does nothing. Users are expected to call :meth:`~hunter.tracer.Tracer.trace`.

        Returns: self
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Wrapper around :meth:`~hunter.tracer.Tracer.stop`. Does nothing with the arguments.
        """
        self.stop()



cimport cython


ctypedef extern FrameType

cdef extern from *:
    ctypedef extern class types.CodeType[object PyCodeObject, check_size ignore]:
        cdef object co_filename
        cdef object co_name
        cdef int co_argcount

cdef extern from "vendor/_compat.h":
    """
    static inline PyCodeObject* Hunter_PyFrame_GetCode(PyObject* frame) {
        return PyFrame_GetCode((PyFrameObject*) frame);
    }
    static inline int Hunter_PyFrame_GetLasti(PyObject* frame) {
        return PyFrame_GetLasti((PyFrameObject*) frame);
    }
    static inline int Hunter_PyFrame_GetLineNumber(PyObject* frame) {
        return PyFrame_GetLineNumber((PyFrameObject*) frame);
    }
    static inline PyObject* Hunter_PyFrame_GetGlobals(PyObject* frame) {
        return PyFrame_GetGlobals((PyFrameObject*) frame);
    }
    static inline PyObject* Hunter_PyFrame_GetLocals(PyObject* frame) {
        return PyFrame_GetLocals((PyFrameObject*) frame);
    }
    """
    object PyCode_GetCode(CodeType)
    object PyCode_GetVarnames(CodeType)
    CodeType Hunter_PyFrame_GetCode(FrameType frame)
    int Hunter_PyFrame_GetLasti(FrameType frame)
    int Hunter_PyFrame_GetLineNumber(FrameType frame)
    object Hunter_PyFrame_GetGlobals(FrameType frame)
    object Hunter_PyFrame_GetLocals(FrameType frame)


@cython.final
cdef class Event:
    cdef:
        readonly FrameType frame
        readonly str kind
        readonly object arg
        readonly int depth
        readonly int calls
        readonly bint threading_support
        readonly bint detached
        readonly object builtin

        object _code
        object _filename
        object _fullsource
        object _function
        object _function_object
        object _globals
        object _lineno
        object _locals
        object _module
        object _source
        object _stdlib
        object _thread
        object _threadidn  # slightly different name cause "_threadid" is a goddamn macro in Microsoft stddef.h
        object _threadname
        object _instruction

        CodeType code_getter(self)
        object filename_getter(self)
        object fullsource_getter(self)
        object function_getter(self)
        object globals_getter(self)
        object lineno_getter(self)
        object locals_getter(self)
        object module_getter(self)
        object source_getter(self)
        object stdlib_getter(self)
        object threadid_getter(self)
        object threadname_getter(self)
        object instruction_getter(self)

cdef Event fast_clone(Event self)
cdef Event fast_detach(Event self, object value_filter)



import linecache
from functools import partial
from os.path import basename
from os.path import exists
from os.path import splitext
from threading import current_thread
from tokenize import TokenError
from tokenize import generate_tokens

from .const import SITE_PACKAGES_PATHS
from .const import SYS_PREFIX_PATHS
from .util import CYTHON_SUFFIX_RE
from .util import LEADING_WHITESPACE_RE
from .util import cached_property
from .util import get_func_in_mro
from .util import get_main_thread
from .util import if_same_code

__all__ = ('Event',)


class Event:
    """
    A wrapper object for Frame objects. Instances of this are passed to your custom functions or predicates.

    Provides few convenience properties.

    Args:
        frame (Frame): A python `Frame <https://docs.python.org/3/reference/datamodel.html#frame-objects>`_ object.
        kind (str): A string like ``'call'``, ``'line'``, ``'return'`` or ``'exception'``.
        arg: A value that depends on ``kind``. Usually is ``None`` but for ``'return'`` or ``'exception'`` other values
            may be expected.
        tracer (:class:`hunter.tracer.Tracer`): The :class:`~hunter.tracer.Tracer` instance that created the event.
            Needed for the ``calls`` and ``depth`` fields.
    """

    frame = None
    kind = None
    arg = None
    depth = None
    calls = None
    builtin = None

    def __init__(
        self,
        frame,
        kind,
        arg,
        depth,
        calls,
        threading_support,
    ):
        #: The original Frame object.
        #:
        #: .. note::
        #:
        #:  Not allowed in the builtin predicates (it's the actual Frame object).
        #:  You may access it from your custom predicate though.
        self.frame = frame

        if kind.startswith('c_'):
            kind = kind[2:]
            builtin = True
        else:
            builtin = False

        #: If kind of the event is one of ``'c_call'``, ``'c_return'``, or ``'c_exception'`` then this will be True.
        #:
        #: :type: bool
        self.builtin = builtin

        #: The kind of the event, could be one of ``'call'``, ``'line'``, ``'return'``, ``'exception'``.
        #:
        #: :type: str
        self.kind = kind

        #: A value that depends on ``kind``
        self.arg = arg

        #: Tracing depth (increases on calls, decreases on returns).
        #:
        #: :type: int
        self.depth = depth

        #: A counter for total number of calls up to this Event.
        #:
        #: :type: int
        self.calls = calls

        #: A copy of the :attr:`hunter.tracer.Tracer.threading_support` flag.
        #:
        #: .. note::
        #:
        #:  Not allowed in the builtin predicates. You may access it from your custom predicate though.
        #:
        #: :type: bool or None
        self.threading_support = threading_support

        #: Flag that is ``True`` if the event was created with :meth:`~hunter.event.Event.detach`.
        #:
        #: :type: bool
        self.detached = False

    def __repr__(self):
        return (
            f'<Event kind={self.kind!r} function={self.function!r} module={self.module!r} filename={self.filename!r} lineno={self.lineno}>'
        )

    def __eq__(self, other):
        return self is other

    def __reduce__(self):
        raise TypeError("cannot pickle 'hunter.event.Event' object")

    def detach(self, value_filter=None):
        """
        Return a copy of the event with references to live objects (like the frame) removed. You should use this if you
        want to store or use the event outside the handler.

        You should use this if you want to avoid memory leaks or side-effects when storing the events.

        Args:
            value_filter:
                Optional callable that takes one argument: ``value``.

                If not specified then the ``arg``, ``globals`` and ``locals`` fields will be ``None``.

        Example usage in a :class:`~hunter.actions.ColorStreamAction` subclass:

        .. sourcecode:: python

            def __call__(self, event):
                self.events = [event.detach(lambda field, value: self.try_repr(value))]

        """
        event = Event.__new__(Event)

        event.__dict__['code'] = self.code
        event.__dict__['filename'] = self.filename
        event.__dict__['fullsource'] = self.fullsource
        event.__dict__['function'] = self.function
        event.__dict__['lineno'] = self.lineno
        event.__dict__['module'] = self.module
        event.__dict__['source'] = self.source
        event.__dict__['stdlib'] = self.stdlib
        event.__dict__['threadid'] = self.threadid
        event.__dict__['threadname'] = self.threadname
        event.__dict__['instruction'] = self.instruction

        if value_filter:
            event.__dict__['arg'] = value_filter(self.arg)
            event.__dict__['globals'] = {key: value_filter(value) for key, value in self.globals.items()}
            event.__dict__['locals'] = {key: value_filter(value) for key, value in self.locals.items()}
        else:
            event.__dict__['globals'] = {}
            event.__dict__['locals'] = {}
            event.__dict__['arg'] = None

        event.threading_support = self.threading_support
        event.calls = self.calls
        event.depth = self.depth
        event.kind = self.kind
        event.builtin = self.builtin

        event.detached = True

        return event

    def clone(self):
        event = Event.__new__(Event)
        event.__dict__ = dict(self.__dict__)
        return event

    @cached_property
    def instruction(self):
        """
        Last byte instruction. If no bytecode was used (Cython code) then it returns ``None``.
        Depending on Python version it might be an int or a single char string.

        :type: int or single char string or None
        """
        if self.frame.f_code.co_code and len(self.frame.f_code.co_code) > self.frame.f_lasti >= 0:
            return self.frame.f_code.co_code[self.frame.f_lasti]

    @cached_property
    def threadid(self):
        """
        Current thread ident. If current thread is main thread then it returns ``None``.

        :type: int or None
        """
        current = self._thread.ident
        main = get_main_thread()
        if main is None:
            return current
        else:
            return current if current != main.ident else None

    @cached_property
    def threadname(self):
        """
        Current thread name.

        :type: str
        """
        return self._thread.name

    @cached_property
    def _thread(self):
        return current_thread()

    @cached_property
    def locals(self):
        """
        A dict with local variables.

        :type: dict
        """
        if self.builtin:
            return {}

        return self.frame.f_locals

    @cached_property
    def globals(self):
        """
        A dict with global variables.

        :type: dict
        """
        if self.builtin:
            return {}

        return self.frame.f_globals

    @cached_property
    def function(self):
        """
        A string with function name.

        :type: str
        """
        if self.builtin:
            return self.arg.__name__
        else:
            return self.code.co_name

    @cached_property
    def function_object(self):
        """
        The function instance.

        .. warning:: Use with prudence.

            * Will be ``None`` for decorated functions on Python 2 (methods may still work tho).
            * May be ``None`` if tracing functions or classes not defined at module level.
            * May be very slow if tracing modules with lots of variables.

        :type: function or None
        """
        # Based on MonkeyType's get_func
        if self.builtin:
            return self.builtin

        code = self.code
        if code.co_name is None:
            return None
        # First, try to find the function in globals
        candidate = self.globals.get(code.co_name, None)
        func = if_same_code(candidate, code)
        # If that failed, as will be the case with class and instance methods, try
        # to look up the function from the first argument. In the case of class/instance
        # methods, this should be the class (or an instance of the class) on which our
        # method is defined.
        if func is None and code.co_argcount >= 1:
            first_arg = self.locals.get(code.co_varnames[0])
            func = get_func_in_mro(first_arg, code)
        # If we still can't find the function, as will be the case with static methods,
        # try looking at classes in global scope.
        if func is None:
            for v in self.globals.values():
                if not isinstance(v, type):
                    continue
                func = get_func_in_mro(v, code)
                if func is not None:
                    break
        return func

    @cached_property
    def module(self):
        """
        A string with module name (like ``'foo.bar'``).

        :type: str
        """
        if self.builtin:
            module = self.arg.__module__
        else:
            module = self.frame.f_globals.get('__name__', '')
        if module is None:
            module = '?'
        return module

    @cached_property
    def filename(self):
        """
        A string with the path to the module's file. May be empty if ``__file__`` attribute is missing.
        May be relative if running scripts.

        :type: str
        """
        # if self.builtin:
        #     return '<builtin>'
        # if self.builtin:
        #     return '<builtin>'
        filename = self.frame.f_code.co_filename
        if not filename:
            filename = self.frame.f_globals.get('__file__')
        if not filename:
            filename = '?'
        if filename.endswith(('.pyc', '.pyo')):
            filename = filename[:-1]
        elif filename.endswith('$py.class'):  # Jython
            filename = filename[:-9] + '.py'
        elif filename.endswith(('.so', '.pyd')):
            basename = CYTHON_SUFFIX_RE.sub('', filename)
            for ext in ('.pyx', '.py'):
                cyfilename = basename + ext
                if exists(cyfilename):
                    filename = cyfilename
                    break
        return filename

    @cached_property
    def lineno(self):
        """
        An integer with line number in file.

        :type: int
        """
        return self.frame.f_lineno

    @cached_property
    def code(self):
        """
        A code object (not a string).
        """
        return self.frame.f_code

    @cached_property
    def stdlib(self):
        """
        A boolean flag. ``True`` if frame is in stdlib.

        :type: bool
        """
        module_parts = self.module.split('.')
        if 'pkg_resources' in module_parts:
            # skip this over-vendored module
            return True
        elif self.filename == '<string>' and (self.module.startswith('namedtuple_') or self.module == 'site'):
            # skip namedtuple exec garbage
            return True
        elif self.filename.startswith(SITE_PACKAGES_PATHS):
            # if it's in site-packages then its definitely not stdlib
            return False
        elif self.filename.startswith(SYS_PREFIX_PATHS):
            return True
        else:
            return False

    @cached_property
    def fullsource(self):
        """
        A string with the sourcecode for the current statement (from ``linecache`` - failures are ignored).

        May include multiple lines if it's a class/function definition (will include decorators).

        :type: str
        """
        try:
            if self.kind == 'call' and self.code.co_name != '<module>':
                lines = []
                try:
                    for _, token, _, _, _ in generate_tokens(
                        partial(
                            next,
                            yield_lines(
                                self.filename,
                                self.frame.f_globals,
                                self.lineno - 1,
                                lines.append,
                            ),
                        )
                    ):
                        if token in ('def', 'class', 'lambda'):
                            return ''.join(lines)
                except TokenError:
                    pass

            return linecache.getline(self.filename, self.lineno, self.frame.f_globals)
        except Exception as exc:
            return f'??? NO SOURCE: {exc!r}'

    @cached_property
    def source(self):
        """
        A string with the sourcecode for the current line (from ``linecache`` - failures are ignored).

        Fast but sometimes incomplete.

        :type: str
        """
        if self.filename.endswith(('.so', '.pyd')):
            return f'??? NO SOURCE: not reading binary {splitext(basename(self.filename))[1]} file'
        try:
            return linecache.getline(self.filename, self.lineno, self.frame.f_globals)
        except Exception as exc:
            return f'??? NO SOURCE: {exc!r}'

    __getitem__ = object.__getattribute__


def yield_lines(
    filename,
    module_globals,
    start,
    collector,
    limit=10,
    leading_whitespace_re=LEADING_WHITESPACE_RE,
):
    dedent = None
    amount = 0
    for line in linecache.getlines(filename, module_globals)[start : start + limit]:
        if dedent is None:
            dedent = leading_whitespace_re.findall(line)
            dedent = dedent[0] if dedent else ''
            amount = len(dedent)
        elif not line.startswith(dedent):
            break
        collector(line)
        yield line[amount:]



# cython: linetrace=True, language_level=3str, c_string_encoding=ascii, freethreading_compatible=True
cimport cython

from functools import partial
from linecache import getline
from linecache import getlines
from os.path import basename
from os.path import exists
from os.path import splitext
from threading import current_thread
from tokenize import TokenError
from tokenize import generate_tokens

from cpython.pythread cimport PyThread_get_thread_ident
from cpython.ref cimport PyObject

from .const import SITE_PACKAGES_PATHS
from .const import SYS_PREFIX_PATHS
from .util import CYTHON_SUFFIX_RE
from .util import LEADING_WHITESPACE_RE
from .util import get_func_in_mro
from .util import get_main_thread
from .util import if_same_code

__all__ = 'Event',

cdef object UNSET = object()


cdef str CALL = 'call'
cdef str EXCEPTION = 'exception'
cdef str LINE = 'line'
cdef str RETURN = 'return'

cdef const PyObject** KIND_NAMES = [
    <PyObject*>CALL,
    <PyObject*>EXCEPTION,
    <PyObject*>LINE,
    <PyObject*>RETURN,
    <PyObject*>CALL,
    <PyObject*>EXCEPTION,
    <PyObject*>RETURN,
]


@cython.auto_pickle(False)
cdef class Event:
    """
    A wrapper object for Frame objects. Instances of this are passed to your custom functions or predicates.

    Provides few convenience properties.

    Args:
        frame (Frame): A python `Frame <https://docs.python.org/3/reference/datamodel.html#frame-objects>`_ object.
        kind (str): A string like ``'call'``, ``'line'``, ``'return'`` or ``'exception'``.
        arg: A value that depends on ``kind``. Usually is ``None`` but for ``'return'`` or ``'exception'`` other values
            may be expected.
    """
    def __init__(self, FrameType frame, int kind, object arg, int depth, int calls, bint threading_support):
        self.arg = arg
        self.frame = frame
        self.kind = <str> KIND_NAMES[kind]
        self.depth = depth
        self.calls = calls
        self.threading_support = threading_support
        self.detached = False
        self.builtin = kind > 3

        self._code = UNSET
        self._filename = UNSET
        self._fullsource = UNSET
        self._function_object = UNSET
        self._function = UNSET
        self._globals = UNSET
        self._lineno = UNSET
        self._locals = UNSET
        self._module = UNSET
        self._source = UNSET
        self._stdlib = UNSET
        self._threadidn = UNSET
        self._threadname = UNSET
        self._thread = UNSET
        self._instruction = UNSET

    def __repr__(self):
        return '<Event kind=%r function=%r module=%r filename=%r lineno=%s>' % (
            self.kind, self.function, self.module, self.filename, self.lineno
        )

    def __eq__(self, other):
        return self is other

    def detach(self, value_filter=None):
        return fast_detach(self, value_filter)

    def clone(self):
        return fast_clone(self)

    cdef inline instruction_getter(self):
        cdef int position

        if self._instruction is UNSET:
            position = Hunter_PyFrame_GetLasti(self.frame)
            co_code = PyCode_GetCode(self.code_getter())
            if co_code and len(co_code) > position >= 0:
                self._instruction = co_code[position]
            else:
                self._instruction = None
        return self._instruction

    @property
    def instruction(self):
        return self.instruction_getter()

    cdef inline threadid_getter(self):
        cdef long current

        if self._threadidn is UNSET:
            current = PyThread_get_thread_ident()
            main = get_main_thread()
            if main is not None and current == main.ident:
                self._threadidn = None
            else:
                self._threadidn = current
        return self._threadidn

    @property
    def threadid(self):
        return self.threadid_getter()

    cdef inline threadname_getter(self):
        if self._threadname is UNSET:
            if self._thread is UNSET:
                self._thread = current_thread()
            self._threadname = self._thread.name
        return self._threadname

    @property
    def threadname(self):
        return self.threadname_getter()

    cdef inline locals_getter(self):
        if self._locals is UNSET:
            if self.builtin:
                self._locals = {}
            else:
                self._locals = Hunter_PyFrame_GetLocals(self.frame)
        return self._locals

    @property
    def locals(self):
        return self.locals_getter()

    cdef inline globals_getter(self):
        if self._globals is UNSET:
            if self.builtin:
                self._locals = {}
            else:
                self._globals = Hunter_PyFrame_GetGlobals(self.frame)
        return self._globals

    @property
    def globals(self):
        return self.globals_getter()

    cdef inline function_getter(self):
        if self._function is UNSET:
            if self.builtin:
                self._function = self.arg.__name__
            else:
                self._function = self.code_getter().co_name
        return self._function

    @property
    def function(self):
        return self.function_getter()

    @property
    def function_object(self):
        cdef CodeType code
        if self.builtin:
            return self.builtin
        elif self._function_object is UNSET:
            code = self.code_getter()
            if code.co_name is None:
                return None
            # First, try to find the function in globals
            candidate = self.globals.get(code.co_name, None)
            func = if_same_code(candidate, code)
            # If that failed, as will be the case with class and instance methods, try
            # to look up the function from the first argument. In the case of class/instance
            # methods, this should be the class (or an instance of the class) on which our
            # method is defined.
            if func is None and code.co_argcount >= 1:
                first_arg = self.locals.get(PyCode_GetVarnames(code)[0])
                func = get_func_in_mro(first_arg, code)
            # If we still can't find the function, as will be the case with static methods,
            # try looking at classes in global scope.
            if func is None:
                for v in self.globals.values():
                    if not isinstance(v, type):
                        continue
                    func = get_func_in_mro(v, code)
                    if func is not None:
                        break
            self._function_object = func
        return self._function_object

    cdef inline module_getter(self):
        if self._module is UNSET:
            if self.builtin:
                module = self.arg.__module__
            else:
                module = self.globals.get('__name__', '')
            if module is None:
                module = '?'
            self._module = module
        return self._module

    @property
    def module(self):
        return self.module_getter()

    cdef inline filename_getter(self):
        cdef CodeType code
        if self._filename is UNSET:
            code = self.code_getter()
            filename = code.co_filename
            if not filename:
                filename = self.globals.get('__file__')
            if not filename:
                filename = '?'
            elif filename.endswith(('.pyc', '.pyo')):
                filename = filename[:-1]
            elif filename.endswith(('.so', '.pyd')):
                cybasename = CYTHON_SUFFIX_RE.sub('', filename)
                for ext in ('.pyx', '.py'):
                    cyfilename = cybasename + ext
                    if exists(cyfilename):
                        filename = cyfilename
                        break

            self._filename = filename
        return self._filename

    @property
    def filename(self):
        return self.filename_getter()

    cdef inline lineno_getter(self):
        if self._lineno is UNSET:
            self._lineno = Hunter_PyFrame_GetLineNumber(self.frame)
        return self._lineno

    @property
    def lineno(self):
        return self.lineno_getter()

    cdef inline CodeType code_getter(self):
        if self._code is UNSET:
            return Hunter_PyFrame_GetCode(self.frame)
        else:
            return self._code

    @property
    def code(self):
        return self.code_getter()

    cdef inline stdlib_getter(self):
        if self._stdlib is UNSET:
            module_parts = self.module.split('.')
            if 'pkg_resources' in module_parts:
                # skip this over-vendored module
                self._stdlib = True
            elif self.filename == '<string>' and (self.module.startswith('namedtuple_') or self.module == 'site'):
                # skip namedtuple exec garbage
                self._stdlib = True
            elif self.filename.startswith(SITE_PACKAGES_PATHS):
                # if in site-packages then definitely not stdlib
                self._stdlib = False
            elif self.filename.startswith(SYS_PREFIX_PATHS):
                self._stdlib = True
            else:
                self._stdlib = False
        return self._stdlib

    @property
    def stdlib(self):
        return self.stdlib_getter()

    cdef inline fullsource_getter(self):
        cdef list lines
        cdef CodeType code

        if self._fullsource is UNSET:
            try:
                self._fullsource = None
                code = self.code_getter()
                if self.kind == 'call' and code.co_name != '<module>':
                    lines = []
                    try:
                        for _, token, _, _, _ in generate_tokens(
                            partial(
                                next,
                                yield_lines(
                                    self.filename,
                                    self.frame.f_globals,
                                    self.lineno - 1,
                                    lines,
                                ),
                            )
                        ):
                            if token in ('def', 'class', 'lambda'):
                                self._fullsource = ''.join(lines)
                                break
                    except TokenError:
                        pass
                if self._fullsource is None:
                    self._fullsource = getline(self.filename, self.lineno, self.globals)
            except Exception as exc:
                self._fullsource = f'??? NO SOURCE: {exc!r}'
        return self._fullsource

    @property
    def fullsource(self):
        return self.fullsource_getter()

    cdef inline source_getter(self):
        if self._source is UNSET:
            if self.filename.endswith(('.so', '.pyd')):
                self._source = f'??? NO SOURCE: not reading binary {splitext(basename(self.filename))[1]} file'
            try:
                self._source = getline(self.filename, self.lineno, self.globals)
            except Exception as exc:
                self._source = f'??? NO SOURCE: {exc!r}'

        return self._source

    @property
    def source(self):
        return self.source_getter()

    def __getitem__(self, item):
        return getattr(self, item)


def yield_lines(filename, module_globals, start, list collector,
                limit=10):
    dedent = None
    amount = 0
    for line in getlines(filename, module_globals)[start:start + limit]:
        if dedent is None:
            dedent = LEADING_WHITESPACE_RE.findall(line)
            dedent = dedent[0] if dedent else ''
            amount = len(dedent)
        elif not line.startswith(dedent):
            break
        collector.append(line)
        yield line[amount:]


cdef inline Event fast_detach(Event self, object value_filter):
    event = <Event> Event.__new__(Event)

    event._code = self.code_getter()
    event._filename = self.filename_getter()
    event._fullsource = self.fullsource_getter()
    event._function_object = self._function_object
    event._function = self.function_getter()
    event._lineno = self.lineno_getter()
    event._module = self.module_getter()
    event._source = self.source_getter()
    event._stdlib = self.stdlib_getter()
    event._threadidn = self.threadid_getter()
    event._threadname = self.threadname_getter()
    event._instruction = self.instruction_getter()

    if value_filter:
        event._globals = {key: value_filter(value) for key, value in self.globals.items()}
        event._locals = {key: value_filter(value) for key, value in self.locals.items()}
        event.arg = value_filter(self.arg)
    else:
        event._globals = {}
        event._locals = {}
        event.arg = None

    event.builtin = self.builtin
    event.calls = self.calls
    event.depth = self.depth
    event.detached = True
    event.kind = self.kind
    event.threading_support = self.threading_support

    return event

cdef inline Event fast_clone(Event self):
    event = <Event> Event.__new__(Event)
    event.arg = self.arg
    event.builtin = self.builtin
    event.calls = self.calls
    event.depth = self.depth
    event.detached = False
    event.frame = self.frame
    event.kind = self.kind
    event.threading_support = self.threading_support
    event._code = self._code
    event._filename = self._filename
    event._fullsource = self._fullsource
    event._function_object = self._function_object
    event._function = self._function
    event._globals = self._globals
    event._lineno = self._lineno
    event._locals = self._locals
    event._module = self._module
    event._source = self._source
    event._stdlib = self._stdlib
    event._threadidn = self._threadidn
    event._threadname = self._threadname
    event._thread = self._thread
    event._instruction = self._instruction
    return event



import collections
import os
import site
import stat
import sys
import sysconfig

SITE_PACKAGES_PATHS = set()
for scheme in sysconfig.get_scheme_names():
    if scheme == 'posix_home':
        # it would appear this scheme is not for site-packages
        continue
    for name in ['platlib', 'purelib']:
        try:
            SITE_PACKAGES_PATHS.add(sysconfig.get_path(name, scheme))
        except KeyError:
            pass
if hasattr(site, 'getusersitepackages'):
    SITE_PACKAGES_PATHS.add(site.getusersitepackages())
if sys.version_info < (3, 10):
    from distutils.sysconfig import get_python_lib

    SITE_PACKAGES_PATHS.add(get_python_lib())
SITE_PACKAGES_PATHS.add(os.path.dirname(os.path.dirname(__file__)))
SITE_PACKAGES_PATHS = tuple(SITE_PACKAGES_PATHS)

SYS_PREFIX_PATHS = {
    '<frozen zipimport>',
    '<frozen importlib._bootstrap>',
    '<frozen importlib._bootstrap_external>',
    sys.prefix,
    sys.exec_prefix,
    os.path.dirname(os.__file__),
    os.path.dirname(stat.__file__),
    os.path.dirname(collections.__file__),
}
for prop in (
    'real_prefix',
    'real_exec_prefix',
    'base_prefix',
    'base_exec_prefix',
):
    if hasattr(sys, prop):
        SYS_PREFIX_PATHS.add(getattr(sys, prop))

SYS_PREFIX_PATHS = tuple(sorted(SYS_PREFIX_PATHS, key=len, reverse=True))
